# Copyright 2015 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it 
# under the terms of the GNU General Public License version 3, as published 
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but 
# WITHOUT ANY WARRANTY; without even the implied warranties of 
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR 
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along 
# with this program.  If not, see <http://www.gnu.org/licenses/>.

import email.utils
import errno
import logging
import os.path
import shutil
import subprocess

from debian import (
    changelog,
    deb822,
    )

from gitbuildrecipe.deb_version import substitute_changelog_vars
from gitbuildrecipe.recipe import SubstitutionUnavailable


class MissingDependency(Exception):
    pass


class NoSuchTag(Exception):

    def __init__(self, tag_name):
        super().__init__()
        self.tag_name = tag_name


def debian_source_package_name(control_path):
    """Open a debian control file and extract the package name.

    """
    with open(control_path, "r") as f:
        control = deb822.Deb822(f)
        return control["Source"]


def extract_upstream_tarball(path, package, version, dest_dir):
    """Extract the upstream tarball from a Git repository.

    :param path: Path to the Git repository
    :param package: Package name
    :param version: Package version
    :param dest_dir: Destination directory
    """
    prefix = "%s_%s.orig.tar." % (package, version)
    dest_filename = None
    pristine_tar_list = subprocess.Popen(
        ["pristine-tar", "list"], stdout=subprocess.PIPE, cwd=path)
    try:
        for line in pristine_tar_list.stdout:
            line = line.decode("UTF-8", errors="replace").rstrip("\n")
            if line.startswith(prefix):
                dest_filename = line
    finally:
        pristine_tar_list.wait()
    if dest_filename is not None:
        subprocess.check_call(
            ["pristine-tar", "checkout",
             os.path.abspath(os.path.join(dest_dir, dest_filename))],
            cwd=path)
    else:
        tag_names = ["upstream/%s" % version, "upstream-%s" % version]
        git_tag_list = subprocess.Popen(
            ["git", "tag"], stdout=subprocess.PIPE, cwd=path)
        try:
            for line in git_tag_list.stdout:
                line = line.decode("UTF-8", errors="replace").rstrip("\n")
                if line in tag_names:
                    tag = line
                    break
            else:
                raise NoSuchTag(tag_names[0])
        finally:
            git_tag_list.wait()
        # Default to .tar.gz
        dest_filename = prefix + "gz"
        with open(os.path.join(dest_dir, dest_filename), "wb") as dest:
            subprocess.check_call(
                ["git", "archive", "--format=tar.gz",
                 "--prefix=%s-%s/" % (package, version), tag],
                stdout=dest, cwd=path)


def add_autobuild_changelog_entry(base_branch, basedir, package,
                                  distribution=None, author_name=None,
                                  author_email=None, append_version=None):
    """Add a new changelog entry for an autobuild.

    :param base_branch: Recipe base branch
    :param basedir: Base working directory
    :param package: package name
    :param distribution: Optional distribution (defaults to last entry
        distribution)
    :param author_name: Name of the build requester
    :param author_email: Email of the build requester
    :param append_version: Optional version suffix to add
    """
    debian_dir = os.path.join(basedir, "debian")
    if not os.path.exists(debian_dir):
        os.makedirs(debian_dir)
    cl_path = os.path.join(debian_dir, "changelog")
    file_found = os.path.exists(cl_path)
    if file_found:
        with open(cl_path) as cl_f:
            contents = cl_f.read()
        cl = changelog.Changelog(file=contents)
    else:
        cl = changelog.Changelog()
    if len(cl._blocks) > 0:
        if distribution is None:
            distribution = cl._blocks[0].distributions.split()[0]
    else:
        if file_found:
            if len(contents.strip()) > 0:
                reason = (
                    "debian/changelog didn't contain any parseable stanzas")
            else:
                reason = "debian/changelog was empty"
        else:
            reason = "debian/changelog was not present"
        if distribution is None:
            distribution = "unstable"
    if base_branch.format_version in (0.1, 0.2, 0.3, 0.4):
        try:
            substitute_changelog_vars(base_branch, None, cl)
        except SubstitutionUnavailable as e:
            raise Exception(
                "No previous changelog to take the upstream version from as "
                "%s was used: %s: %s." % (e.name, e.reason, reason))
    # Use debian packaging environment variables
    # or default values if they don't exist
    if author_name is None or author_email is None:
        author_name, author_email = changelog.get_maintainer()
    author = "%s <%s>" % (author_name, author_email)

    date = email.utils.formatdate(localtime=True)
    version = base_branch.deb_version
    if append_version is not None:
        version += append_version
    try:
        changelog.Version(version)
    except (changelog.VersionError, ValueError) as e:
        raise Exception("Invalid deb-version: %s: %s" % (version, e))
    cl.new_block(package=package, version=version,
            distributions=distribution, urgency="low",
            changes=["", "  * Auto build.", ""],
            author=author, date=date)
    with open(cl_path, "w") as cl_f:
        cl.write_to_open_file(cl_f)


def build_source_package(basedir, tgz_check=True):
    command = ["/usr/bin/debuild"]
    if tgz_check:
        command.append("--tgz-check")
    else:
        command.append("--no-tgz-check")
    command.extend(["-i", "-I", "-S", "-uc", "-us"])
    logging.info("Building the source package")
    try:
        subprocess.check_call(command, cwd=basedir)
    except OSError as e:
        if e.errno == errno.ENOENT:
            raise MissingDependency(
                "debuild is not installed; please install the devscripts "
                "package.")
        raise


def get_source_format(path):
    """Retrieve the source format name from a package.

    :param path: Path to the package
    :return: String with package format
    """
    source_format_path = os.path.join(path, "debian", "source", "format")
    if not os.path.exists(source_format_path):
        return "1.0"
    with open(source_format_path) as f:
        return f.read().strip()


def convert_3_0_quilt_to_native(path):
    """Convert a package in 3.0 (quilt) format to 3.0 (native).

    This applies all patches in the package and updates the
    debian/source/format file.

    :param path: Path to the package on disk
    """
    path = os.path.abspath(path)
    patches_dir = os.path.join(path, "debian", "patches")
    series_file = os.path.join(patches_dir, "series")
    if os.path.exists(series_file):
        logging.info("Applying quilt patches")
        env = dict(os.environ)
        env["QUILT_SERIES"] = series_file
        env["QUILT_PATCHES"] = patches_dir
        try:
            subprocess.check_call(
                ["quilt", "push", "-a", "-v"], cwd=path, env=env)
        except OSError as e:
            if e.errno == errno.ENOENT:
                raise MissingDependency(
                    "quilt is not installed; please install it.")
            raise
        except subprocess.CalledProcessError as e:
            if e.returncode != 2:
                raise Exception("Failed to apply quilt patches")
    if os.path.exists(patches_dir):
        shutil.rmtree(patches_dir)
    with open(os.path.join(path, "debian", "source", "format"), "w") as f:
        f.write("3.0 (native)\n")


def force_native_format(working_tree_path, current_format):
    """Make sure a package is a format that supports native packages.

    :param working_tree_path: Path to the package
    """
    if current_format == "3.0 (quilt)":
        convert_3_0_quilt_to_native(working_tree_path)
    elif current_format not in ("1.0", "3.0 (native)"):
        raise Exception("Unknown source format %s" % current_format)



#! /usr/bin/python3

# Copyright 2015 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it 
# under the terms of the GNU General Public License version 3, as published 
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but 
# WITHOUT ANY WARRANTY; without even the implied warranties of 
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR 
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along 
# with this program.  If not, see <http://www.gnu.org/licenses/>.

"""A Git command to construct trees based on recipes."""

import argparse
import datetime
import os.path
import shutil
import sys
import tempfile

from debian.changelog import Changelog

from gitbuildrecipe.deb_util import (
    NoSuchTag,
    add_autobuild_changelog_entry,
    build_source_package,
    debian_source_package_name,
    extract_upstream_tarball,
    force_native_format,
    get_source_format,
    )
from gitbuildrecipe.deb_version import (
    check_expanded_deb_version,
    substitute_branch_vars,
    substitute_time,
    )
from gitbuildrecipe.recipe import (
    build_tree,
    parse_recipe,
    resolve_revisions,
    )


def main():
    parser = argparse.ArgumentParser(
        description="Construct a Debian source tree based on a recipe.")
    parser.add_argument(
        "location", metavar="LOCATION", type=argparse.FileType("r"),
        help="The file system path to the recipe.")
    parser.add_argument(
        "working_basedir", metavar="WORKING-BASEDIR", nargs="?", help=(
            "The path to a working directory.  If not specified, use a "
            "temporary directory."))
    parser.add_argument(
        "--manifest", metavar="PATH", help="Path to write the manifest to.")
    parser.add_argument(
        "--if-changed-from", metavar="PATH", help=(
            "Only build if the outcome would be different to that "
            "specified in this manifest."))
    parser.add_argument(
        "--package", help=(
            "The package name to use in the changelog entry.  If not "
            "specified then the package from the previous changelog entry "
            "will be used, so it must be specified if there is no changelog."))
    parser.add_argument(
        "--distribution", help=(
            "The distribution to target.  If not specified then the same "
            "distribution as the last entry in debian/changelog will be "
            "used."))
    parser.add_argument(
        "--no-build", action="store_false", default=True, dest="build",
        help="Just ready the source package and don't actually build it.")
    parser.add_argument(
        "--append-version", help=(
            "Append the specified string to the end of the version used in "
            "debian/changelog."))
    parser.add_argument(
        "--safe", action="store_true", default=False,
        help="Error if the recipe would cause arbitrary code execution.")
    parser.add_argument(
        "--allow-fallback-to-native", action="store_true", default=False,
        help=(
            "Allow falling back to a native package if the upstream tarball "
            "cannot be found."))
    args = parser.parse_args()

    base_branch = parse_recipe(args.location, safe=args.safe)
    working_basedir = args.working_basedir
    if working_basedir is None:
        temp_dir = tempfile.mkdtemp(prefix="git-build-recipe-")
        working_basedir = temp_dir
    else:
        temp_dir = None
        if not os.path.exists(working_basedir):
            os.makedirs(working_basedir)
    try:
        package_name = args.package
        if package_name is None:
            package_name = os.path.basename(args.location.name)
            if package_name.endswith(".recipe"):
                package_name = package_name[:-len(".recipe")]
        working_directory = os.path.join(working_basedir, package_name)
        manifest_path = os.path.join(
            working_directory, "debian", "git-build-recipe.manifest")
        build_tree(base_branch, working_directory)
        old_recipe = None
        if args.if_changed_from is not None:
            try:
                if_changed_from = open(args.if_changed_from)
            except FileNotFoundError:
                if_changed_from = None
            if if_changed_from is not None:
                try:
                    old_recipe = parse_recipe(args.if_changed_from)
                finally:
                    if_changed_from.close()
        if base_branch.deb_version is not None:
            time = datetime.datetime.utcnow()
            substitute_time(base_branch, time)
            resolve_revisions(
                base_branch, if_changed_from=old_recipe,
                substitute_branch_vars=substitute_branch_vars)
            check_expanded_deb_version(base_branch)
        else:
            resolve_revisions(base_branch, if_changed_from=old_recipe)
        control_path = os.path.join(working_directory, "debian", "control")
        if not os.path.exists(control_path):
            if args.package is None:
                parser.error(
                    "No control file to take the package name from, and "
                    "--package not specified.")
            package = args.package
        else:
            package = debian_source_package_name(control_path)
        with open(manifest_path, "w") as manifest:
            manifest.write(str(base_branch))
        if base_branch.deb_version is not None:
            # add_autobuild_changelog_entry also substitutes {debupstream}.
            add_autobuild_changelog_entry(
                base_branch, working_directory, package,
                distribution=args.distribution,
                append_version=args.append_version)
        elif args.append_version:
            parser.error(
                "--append-version is only supported for autobuild recipes "
                "(with a 'deb-version' header).")
        changelog_path = os.path.join(working_directory, "debian", "changelog")
        with open(changelog_path) as changelog_file:
            cl = Changelog(file=changelog_file)
        package_name = cl.package
        package_version = cl.version
        current_format = get_source_format(working_directory)
        if (package_version.debian_revision is not None or
                current_format == "3.0 (quilt)"):
            # Non-native package
            try:
                extract_upstream_tarball(
                    base_branch.path, package_name,
                    package_version.upstream_version, working_basedir)
            except NoSuchTag as e:
                if not args.allow_fallback_to_native:
                    parser.error(
                        "Unable to find the upstream source.  Import it as "
                        "tag %s or build with --allow-fallback-to-native." %
                        e.tag_name)
                else:
                    force_native_format(working_directory, current_format)
        if args.build:
            build_source_package(
                working_directory, tgz_check=not args.allow_fallback_to_native)
        if args.manifest is not None:
            with open(args.manifest, "w") as manifest:
                manifest.write(str(base_branch))
    finally:
        if temp_dir is not None:
            shutil.rmtree(temp_dir)


if __name__ == "__main__":
    sys.exit(main())



# Copyright 2015 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it 
# under the terms of the GNU General Public License version 3, as published 
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but 
# WITHOUT ANY WARRANTY; without even the implied warranties of 
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR 
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along 
# with this program.  If not, see <http://www.gnu.org/licenses/>.

import os.path
import subprocess

from fixtures import (
    EnvironmentVariable,
    TempDir,
    )
from testtools import TestCase


class GitRepository:

    def __init__(self, path, allow_create=True):
        self.path = path
        if not os.path.exists(path):
            os.makedirs(path)
        if allow_create and not os.path.exists(os.path.join(path, ".git")):
            self._git_call("init", "-q")

    def _git_call(self, *args, **kwargs):
        subprocess.check_call(["git", "-C", self.path] + list(args), **kwargs)

    def _git_output(self, *args, **kwargs):
        return subprocess.check_output(
            ["git", "-C", self.path] + list(args),
            universal_newlines=True, **kwargs)

    def rev_parse(self, rev):
        return self._git_output("rev-parse", rev).rstrip("\n")

    def last_revision(self):
        return self.rev_parse("HEAD")

    def status(self):
        return self._git_output("status", "--porcelain").splitlines()

    def add(self, files):
        self._git_call("add", *files)

    def commit(self, message, date=None):
        env = os.environ.copy()
        if date is not None:
            env["GIT_COMMITTER_DATE"] = date
        self._git_call("commit", "-q", "--allow-empty", "-m", message, env=env)
        return self.last_revision()

    def tag(self, tag_name, commit, force=False):
        args = ["tag"]
        if force:
            args.append("--force")
        args.extend([tag_name, commit])
        self._git_call(*args, stdout=subprocess.DEVNULL)

    def get_parents(self, commit):
        return self._git_output(
            "log", "-1", "--format=%P", commit).rstrip("\n").split()

    def clone_to(self, new_path):
        subprocess.check_call(["git", "clone", "-q", self.path, new_path])
        return GitRepository(new_path, allow_create=False)

    def build_tree(self, shape):
        """Build a test tree according to a pattern.

        shape is a sequence of file specifications.  If the final character
        is '/', a directory is created.

        This assumes that all the elements in the tree being built are new.

        This doesn't add anything to the index or make any commits.

        :type shape: list or tuple
        :return: None
        """
        if not isinstance(shape, (list, tuple)):
            raise AssertionError(
                "Parameter 'shape' should be a list or a tuple.  Got %r "
                "instead" % (shape,))
        for name in shape:
            assert isinstance(name, str)
            if name.endswith("/"):
                os.mkdir(os.path.join(self.path, name))
            else:
                dirname = os.path.dirname(os.path.join(self.path, name))
                if not os.path.isdir(dirname):
                    os.makedirs(dirname)
                with open(os.path.join(self.path, name), "w") as f:
                    f.write("contents of %s\n" % name)

    def build_tree_contents(self, template):
        """Reconstitute some files from a text description.

        Each element of template is a tuple.  The first element is a
        filename, with an optional ending character indicating the type.

        The template is built relative to the repository's current working
        directory.

        ('foo/',) will build a directory.
        ('foo', 'bar') will write 'bar' to 'foo'.
        ('foo@', 'linktarget') will raise an error.
        """
        for tt in template:
            name = tt[0]
            output_path = os.path.join(self.path, name)
            if name.endswith("/"):
                os.mkdir(output_path)
            elif name.endswith("@"):
                os.symlink(tt[1], output_path[:-1])
            else:
                with open(output_path, "w") as f:
                    f.write(tt[1])


class GitTestCase(TestCase):

    def setUp(self):
        super().setUp()
        # git complains if the user name and email aren't configured at all.
        home = self.make_temporary_directory()
        self.useFixture(EnvironmentVariable("HOME", home))
        with open(os.path.join(home, ".gitconfig"), "w") as gitconfig:
            print("[user]", file=gitconfig)
            print('\tname = "Committer"', file=gitconfig)
            print('\temail = "committer@example.org"', file=gitconfig)

    def make_temporary_directory(self):
        """Create a temporary directory, and return its path."""
        return self.useFixture(TempDir()).path

    def use_temp_dir(self):
        """Use a temporary directory for this test."""
        tempdir = self.make_temporary_directory()
        cwd = os.getcwd()
        os.chdir(tempdir)
        self.addCleanup(os.chdir, cwd)
        return tempdir



# Copyright 2015 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it 
# under the terms of the GNU General Public License version 3, as published 
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but 
# WITHOUT ANY WARRANTY; without even the implied warranties of 
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR 
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along 
# with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Handle deb-version and related substitutions."""

import re
import subprocess

from debian.changelog import Changelog

from gitbuildrecipe.recipe import (
    BranchSubstitutionVariable,
    DateVariable,
    FormattedError,
    GitCommitVariable,
    LatestTagVariable,
    RevdateVariable,
    RevnoVariable,
    RevtimeVariable,
    SubstitutionUnavailable,
    TimeVariable,
    branch_vars,
    simple_vars,
    )


class DebUpstreamVariable(BranchSubstitutionVariable):

    basename = "debupstream"

    minimum_format = 0.1

    def __init__(self, branch_name, version):
        super(DebUpstreamVariable, self).__init__(branch_name)
        self._version = version

    @classmethod
    def from_changelog(cls, branch_name, changelog):
        if len(changelog._blocks) > 0:
            return cls(branch_name, changelog._blocks[0].version)
        else:
            return cls(branch_name, None)

    def get(self):
        if self._version is None:
            raise SubstitutionUnavailable(
                self.name,
                "No previous changelog to take the upstream version from")
        # Should we include the epoch?
        return self._version.upstream_version


class DebVersionVariable(BranchSubstitutionVariable):

    basename = "debversion"

    minimum_format = 0.3

    def __init__(self, branch_name, version):
        super(DebVersionVariable, self).__init__(branch_name)
        self._version = version

    @classmethod
    def from_changelog(cls, branch_name, changelog):
        if len(changelog._blocks) > 0:
            return cls(branch_name, changelog._blocks[0].version)
        else:
            return cls(branch_name, None)

    def get(self):
        if self._version is None:
            raise SubstitutionUnavailable(self.name,
                "No previous changelog to take the version from")
        return str(self._version)


dfsg_regex = re.compile(r"[+.]*dfsg[.]*[0-9]+")

version_regex = re.compile(r"([~+])(svn[0-9]+|bzr[0-9]+|git[0-9a-f]+)")

def version_extract_base(version):
    version = dfsg_regex.sub("", version)
    return version_regex.sub("\\1", version)


class DebUpstreamBaseVariable(DebUpstreamVariable):

    basename = "debupstream-base"
    minimum_format = 0.3

    def get(self):
        version = super(DebUpstreamBaseVariable, self).get()
        version = version_extract_base(version)
        if version[-1] not in ("~", "+"):
            version += "+"
        return version


ok_to_preserve = [
    DebUpstreamVariable,
    DebUpstreamBaseVariable,
    DebVersionVariable,
    ]
deb_branch_vars = [
    DebVersionVariable,
    DebUpstreamBaseVariable,
    DebUpstreamVariable,
    ]


class NotFullyExpanded(FormattedError):

    _fmt = (
        "deb-version not fully expanded: %(deb_version)s.  Valid "
        "substitutions in recipe format %(format_version)s are: "
        "%(available_tokens)s")

    def __init__(self, deb_version, format_version, available_tokens):
        super().__init__(
            deb_version=deb_version, format_version=format_version,
            available_tokens=available_tokens)


def check_expanded_deb_version(base_branch):
    checked_version = base_branch.deb_version
    if checked_version is None:
        return
    for token in ok_to_preserve:
        if issubclass(token, BranchSubstitutionVariable):
            for name in base_branch.list_branch_names():
                checked_version = checked_version.replace(
                    token.determine_name(name), "")
            checked_version = checked_version.replace(
                    token.determine_name(None), "")
        else:
            checked_version = checked_version.replace(
                token.name, "")
    if "{" in checked_version:
        available_tokens = [var.name for var in simple_vars if
                            var.available_in(base_branch.format_version)]
        for var_kls in branch_vars + deb_branch_vars:
            if not var_kls.available_in(base_branch.format_version):
                continue
            for name in base_branch.list_branch_names():
                available_tokens.append(var_kls.determine_name(name))
            available_tokens.append(var_kls.determine_name(None))
        raise NotFullyExpanded(
            base_branch.deb_version, base_branch.format_version,
            available_tokens)


def substitute_changelog_vars(base_branch, branch_name, cl):
    """Substitute variables related to a changelog.

    :param base_branch: The base `RecipeBranch` to operate on.
    :param branch_name: The name of the `RecipeBranch` to substitute (None
        for the root branch).
    :param cl: The `Changelog` object to use.
    """
    debupstream_var = DebUpstreamVariable.from_changelog(branch_name, cl)
    base_branch.deb_version = debupstream_var.replace(base_branch.deb_version)
    if base_branch.format_version < 0.3:
        # The other variables were introduced in recipe format 0.3
        return
    debupstreambase_var = DebUpstreamBaseVariable.from_changelog(
        branch_name, cl)
    base_branch.deb_version = debupstreambase_var.replace(
        base_branch.deb_version)
    pkgversion_var = DebVersionVariable.from_changelog(branch_name, cl)
    base_branch.deb_version = pkgversion_var.replace(base_branch.deb_version)


def substitute_branch_vars(base_branch, child_branch):
    """Substitute the branch variables for the given branch name in deb_version.

    :param base_branch: The base `RecipeBranch` to operate on.
    :param child_branch: The child `RecipeBranch` to operate on.
    """
    if base_branch.deb_version is None:
        return
    revno_var = RevnoVariable(child_branch)
    base_branch.deb_version = revno_var.replace(base_branch.deb_version)
    # XXX assert absence of {svn-revno} for now
    if base_branch.format_version < 0.4:
        # The other variables were introduced in recipe format 0.4
        return
    git_commit_var = GitCommitVariable(child_branch)
    base_branch.deb_version = git_commit_var.replace(base_branch.deb_version)
    latest_tag_var = LatestTagVariable(child_branch)
    base_branch.deb_version = latest_tag_var.replace(base_branch.deb_version)
    revdate_var = RevdateVariable(child_branch)
    base_branch.deb_version = revdate_var.replace(base_branch.deb_version)
    revtime_var = RevtimeVariable(child_branch)
    base_branch.deb_version = revtime_var.replace(base_branch.deb_version)
    changelog_exists = False
    show_process = subprocess.Popen(
        ["git", "-C", child_branch._get_git_path(),
         "show", "%s:debian/changelog" % child_branch.get_revspec()],
        stdout=subprocess.PIPE, stderr=subprocess.DEVNULL)
    try:
        cl = Changelog(file=show_process.stdout)
    finally:
        show_process.stdout.close()
        changelog_exists = (show_process.wait() == 0)
    if changelog_exists:
        substitute_changelog_vars(base_branch, child_branch.name, cl)


def substitute_time(base_branch, time):
    """Substitute the time into deb_version if needed.

    :param time: A `datetime.datetime` with the desired time.
    """
    if base_branch.deb_version is None:
        return
    base_branch.deb_version = TimeVariable(time).replace(
        base_branch.deb_version)
    base_branch.deb_version = DateVariable(time).replace(
        base_branch.deb_version)



# Copyright 2015 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it 
# under the terms of the GNU General Public License version 3, as published 
# by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but 
# WITHOUT ANY WARRANTY; without even the implied warranties of 
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR 
# PURPOSE.  See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along 
# with this program.  If not, see <http://www.gnu.org/licenses/>.

import hashlib
import os.path
import re
import shlex
import signal
import subprocess
import sys
from textwrap import dedent
import unittest

try:
    from debian import changelog
except ImportError:
    from debian_bundle import changelog
from fixtures import EnvironmentVariable
from testtools.content import text_content
from testtools.matchers import (
    EndsWith,
    FileContains,
    MatchesRegex,
    Not,
    PathExists,
    StartsWith,
    )

from gitbuildrecipe.tests import (
    GitRepository,
    GitTestCase,
    )


def skipIfFakeroot():
    return unittest.skipIf(
        "FAKEROOTKEY" in os.environ, "running under fakeroot")


def requirePristineTar():
    path = "/usr/bin/pristine-tar"
    return unittest.skipUnless(os.path.exists(path), "requires %s" % path)


def sha1_file_by_name(name):
    s = hashlib.sha1()
    with open(name, "rb") as f:
        while True:
            buf = f.read(1 << 16)
            if not buf:
                break
            s.update(buf)
    return s.hexdigest()


def make_pristine_tar_delta(dest, tarball_path):
    """Create a pristine-tar delta for a tarball.

    :param dest: Directory to generate pristine tar delta for
    :param tarball_path: Path to the tarball
    :return: pristine-tarball
    """
    def subprocess_setup():
        signal.signal(signal.SIGPIPE, signal.SIG_DFL)
    # If tarball_path is relative, the cwd=dest parameter to Popen will make
    # pristine-tar faaaail. pristine-tar doesn't use the VFS either, so we
    # assume local paths.
    tarball_path = os.path.abspath(tarball_path)
    command = ["pristine-tar", "gendelta", tarball_path, "-"]
    proc = subprocess.Popen(command, stdout=subprocess.PIPE,
            cwd=dest, preexec_fn=subprocess_setup,
            stderr=subprocess.PIPE)
    stdout, stderr = proc.communicate()
    if proc.returncode != 0:
        raise Exception("Generating delta from tar failed: %s" % stderr)
    return stdout


class BlackboxBuilderTests(GitTestCase):

    def setUp(self):
        super().setUp()
        self.use_temp_dir()
        # Replace DEBEMAIL and DEBFULLNAME so that they are known values
        # for the changelog checks.
        self.useFixture(EnvironmentVariable("DEBEMAIL", "maint@maint.org"))
        self.useFixture(EnvironmentVariable("DEBFULLNAME", "M. Maintainer"))

    def _get_file_contents(self, filename, mode="r"):
        """Helper to read contents of a file

        Use FileContains instead to just assert the contents match."""
        self.assertThat(filename, PathExists())
        with open(filename, mode) as f:
            return f.read()

    def run_recipe(self, args, retcode=0):
        cmd = [sys.executable, "-m", "gitbuildrecipe.main"]
        if isinstance(args, str):
            cmd += shlex.split(args)
        else:
            cmd += list(args)
        process = subprocess.Popen(
            cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE,
            universal_newlines=True)
        stdout, stderr = process.communicate()
        if stdout:
            self.addDetail("stdout", text_content(stdout))
        if stderr:
            self.addDetail("stderr", text_content(stderr))
        self.assertEqual(retcode, process.returncode)
        return stdout, stderr

    @skipIfFakeroot()
    def test_cmd_dailydeb(self):
        #TODO: define a test feature for debuild and require it here.
        source = GitRepository("source")
        source.build_tree(["a", "debian/"])
        source.build_tree_contents([
            ("debian/rules", "#!/usr/bin/make -f\nclean:\n"),
            ("debian/control",
                "Source: foo\nMaintainer: maint maint@maint.org\n\n"
                "Package: foo\nArchitecture: all\n")])
        source.add(["a", "debian/rules", "debian/control"])
        commit = source.commit("one")
        with open("test.recipe", "w") as recipe:
            recipe.write(
                "# git-build-recipe format 0.1 deb-version 1\nsource\n")
        out, err = self.run_recipe("--manifest manifest test.recipe working")
        self.assertThat("working/a", Not(PathExists()))
        package_root = "working/test/"
        self.assertThat(os.path.join(package_root, "a"), PathExists())
        self.assertThat(
            os.path.join(package_root, "debian/git-build-recipe.manifest"),
            PathExists())
        self.assertThat("manifest", PathExists())
        self.assertThat("manifest", FileContains(
            "# git-build-recipe format 0.1 deb-version 1\n"
            "source git-commit:%s\n" % commit))
        self.assertThat(
            os.path.join(package_root, "debian/git-build-recipe.manifest"),
            FileContains(
                "# git-build-recipe format 0.1 deb-version 1\n"
                "source git-commit:%s\n" % commit))
        self.assertThat(
            os.path.join(package_root, "debian/changelog"),
            FileContains(matcher=StartsWith(
                "foo (1) unstable; urgency=low\n")))

    @skipIfFakeroot()
    def test_cmd_dailydeb_no_work_dir(self):
        #TODO: define a test feature for debuild and require it here.
        source = GitRepository("source")
        source.build_tree(["a", "debian/"])
        source.build_tree_contents([
            ("debian/rules", "#!/usr/bin/make -f\nclean:\n"),
            ("debian/control",
                "Source: foo\nMaintainer: maint maint@maint.org\n\n"
                "Package: foo\nArchitecture: all\n")])
        source.add(["a", "debian/rules", "debian/control"])
        source.commit("one")
        with open("test.recipe", "w") as recipe:
            recipe.write(
                "# git-build-recipe format 0.1 deb-version 1\nsource\n")
        out, err = self.run_recipe("--manifest manifest test.recipe")

    @skipIfFakeroot()
    def test_cmd_dailydeb_if_changed_from_non_existent(self):
        #TODO: define a test feature for debuild and require it here.
        source = GitRepository("source")
        source.build_tree(["a", "debian/"])
        source.build_tree_contents([
            ("debian/rules", "#!/usr/bin/make -f\nclean:\n"),
            ("debian/control",
                "Source: foo\nMaintainer: maint maint@maint.org\n"
                "\nPackage: foo\nArchitecture: all\n")])
        source.add(["a", "debian/rules", "debian/control"])
        source.commit("one")
        with open("test.recipe", "w") as recipe:
            recipe.write(
                "# git-build-recipe format 0.1 deb-version 1\nsource\n")
        out, err = self.run_recipe(
            "--manifest manifest --if-changed-from bar test.recipe")

    def make_upstream_version(self, source, package_name, version, contents,
                              pristine_tar_format=None):
        source._git_call("checkout", "-q", "--orphan", "upstream")
        source._git_call("rm", "-qrf", ".")
        source.build_tree_contents(contents)
        source.add(["."])
        commit = source.commit("import upstream %s" % version)
        cmd = []
        if pristine_tar_format in (None, "gz"):
            tarfile_path = "%s_%s.orig.tar.gz" % (package_name, version)
        elif pristine_tar_format == "bz2":
            tarfile_path = "%s_%s.orig.tar.bz2" % (package_name, version)
            cmd.extend(["-c", "tar.tar.bz2.command=bzip2 -c"])
        else:
            raise AssertionError(
                "unknown pristine tar format %s" % pristine_tar_format)
        cmd.extend([
            "archive",
            "--prefix=upstream/",
            "-o", os.path.abspath(tarfile_path),
            "upstream",
            ])
        source._git_call(*cmd)
        subprocess.check_call(
            ["pristine-tar", "commit", os.path.abspath(tarfile_path)],
            stderr=subprocess.DEVNULL, cwd=source.path)
        tarfile_sha1 = sha1_file_by_name(tarfile_path)
        source.tag("upstream/%s" % version, commit)
        source._git_call("checkout", "-q", "master")
        return tarfile_sha1

    def make_simple_package(self, path):
        source = GitRepository(path)
        source.build_tree(["a", "debian/"])
        cl_contents = ("package (0.1-1) unstable; urgency=low\n  * foo\n"
                    " -- maint <maint@maint.org>  Tue, 04 Aug 2009 "
                    "10:03:10 +0100\n")
        source.build_tree_contents([
            ("debian/rules", "#!/usr/bin/make -f\nclean:\n"),
            ("debian/control",
                "Source: package\nMaintainer: maint maint@maint.org\n\n"
                "Package: package\nArchitecture: all\n"),
            ("debian/changelog", cl_contents)
            ])
        source.add(["a", "debian/rules", "debian/control", "debian/changelog"])
        source.commit("one")
        return source

    def test_cmd_dailydeb_no_build(self):
        self.make_simple_package("source")
        with open("test.recipe", "w") as recipe:
            recipe.write(
                "# git-build-recipe format 0.1 deb-version 1\nsource\n")
        out, err = self.run_recipe(
            "--manifest manifest --no-build test.recipe working")
        new_cl_contents = (
            "package (1) unstable; urgency=low\n\n"
            "  * Auto build.\n\n -- M. Maintainer <maint@maint.org>  ")
        self.assertThat(
            "working/test/debian/changelog",
            FileContains(matcher=StartsWith(new_cl_contents)))
        for fn in os.listdir("working"):
            self.assertThat(fn, Not(EndsWith(".changes")))

    @skipIfFakeroot()
    def test_cmd_dailydeb_with_package_from_changelog(self):
        #TODO: define a test feature for debuild and require it here.
        self.make_simple_package("source")
        with open("test.recipe", "w") as recipe:
            recipe.write(
                "# git-build-recipe format 0.1 deb-version 1\nsource\n")
        out, err = self.run_recipe(
            "--manifest manifest --if-changed-from bar test.recipe working")
        new_cl_contents = (
            "package (1) unstable; urgency=low\n\n"
            "  * Auto build.\n\n -- M. Maintainer <maint@maint.org>  ")
        self.assertThat(
            "working/test/debian/changelog",
            FileContains(matcher=StartsWith(new_cl_contents)))

    @skipIfFakeroot()
    def test_cmd_dailydeb_with_version_from_changelog(self):
        self.make_simple_package("source")
        with open("test.recipe", "w") as recipe:
            recipe.write(
                "# git-build-recipe format 0.3 deb-version {debversion}-2\n"
                "source\n")
        out, err = self.run_recipe(
            "--allow-fallback-to-native test.recipe working")
        cl = changelog.Changelog(self._get_file_contents(
            "working/test/debian/changelog"))
        self.assertEqual("0.1-1-2", str(cl._blocks[0].version))

    @skipIfFakeroot()
    def test_cmd_dailydeb_with_version_from_other_branch_changelog(self):
        self.make_simple_package("source")
        other = self.make_simple_package("other")
        cl_contents = ("package (0.4-1) unstable; urgency=low\n  * foo\n"
                    " -- maint <maint@maint.org>  Tue, 04 Aug 2009 "
                    "10:03:10 +0100\n")
        other.build_tree_contents([("debian/changelog", cl_contents)])
        other.add(["debian/changelog"])
        other.commit("new changelog entry")
        with open("test.recipe", "w") as recipe:
            recipe.write(
                "# git-build-recipe format 0.4 "
                "deb-version {debversion:other}.2\n"
                "source\n"
                "nest other other other\n")
        out, err = self.run_recipe(
            "--allow-fallback-to-native test.recipe working")
        cl = changelog.Changelog(self._get_file_contents(
            "working/test/debian/changelog"))
        self.assertEqual("0.4-1.2", str(cl._blocks[0].version))

    @skipIfFakeroot()
    def test_cmd_dailydeb_with_upstream_version_from_changelog(self):
        self.make_simple_package("source")
        with open("test.recipe", "w") as recipe:
            recipe.write(
                "# git-build-recipe format 0.1 deb-version {debupstream}-2\n"
                "source\n")
        out, err = self.run_recipe(
            "--allow-fallback-to-native test.recipe working")
        new_cl_contents = (
            "package (0.1-2) unstable; urgency=low\n\n"
            "  * Auto build.\n\n -- M. Maintainer <maint@maint.org>  ")
        self.assertThat(
            "working/test/debian/changelog",
            FileContains(matcher=StartsWith(new_cl_contents)))

    @skipIfFakeroot()
    def test_cmd_dailydeb_with_append_version(self):
        self.make_simple_package("source")
        with open("test.recipe", "w") as recipe:
            recipe.write(
                "# git-build-recipe format 0.1 deb-version 1\nsource\n")
        out, err = self.run_recipe(
            "--append-version ~ppa1 test.recipe working")
        new_cl_contents = (
            "package (1~ppa1) unstable; urgency=low\n\n"
            "  * Auto build.\n\n -- M. Maintainer <maint@maint.org>  ")
        self.assertThat(
            "working/test/debian/changelog",
            FileContains(matcher=StartsWith(new_cl_contents)))

    def test_cmd_dailydeb_with_nonascii_maintainer_in_changelog(self):
        self.useFixture(EnvironmentVariable(
            "DEBFULLNAME", u"Micha\u25c8 Sawicz"))
        self.make_simple_package("source")
        with open("test.recipe", "w") as recipe:
            recipe.write(
                "# git-build-recipe format 0.1 deb-version 1\nsource\n")
        out, err = self.run_recipe("test.recipe working")
        new_cl_contents = (
            "package (1) unstable; urgency=low\n\n"
            "  * Auto build.\n\n"
            u" -- Micha\u25c8 Sawicz <maint@maint.org>  ")
        self.assertThat(
            "working/test/debian/changelog",
            FileContains(matcher=StartsWith(new_cl_contents)))

    def test_cmd_dailydeb_with_invalid_version(self):
        source = GitRepository("source")
        source.build_tree(["a"])
        source.build_tree_contents([
            ("debian/", None),
            ("debian/control",
             "Source: foo\nMaintainer: maint maint@maint.org\n")
            ])
        source.add(["a", "debian/control"])
        source.commit("one")
        with open("test.recipe", "w") as recipe:
            recipe.write(
                "# git-build-recipe format 0.1 deb-version $\nsource\n")
        err = self.run_recipe("test.recipe working", retcode=1)[1]
        self.assertThat(err, MatchesRegex(
            r".*Invalid deb-version: \$: "
            r"(Could not parse version: \$|Invalid version string '\$')\n",
            flags=re.S))

    def test_cmd_dailydeb_with_safe(self):
        self.make_simple_package("source")
        with open("test.recipe", "w") as recipe:
            recipe.write(
                "# git-build-recipe format 0.3 deb-version 1\n"
                "source\nrun something bad")
        out, err = self.run_recipe("--safe test.recipe working", retcode=1)
        self.assertIn("The 'run' instruction is forbidden.\n", err)

    def make_simple_quilt_package(self):
        source = self.make_simple_package("source")
        source.build_tree(["debian/source/"])
        source.build_tree_contents([
            ("debian/source/format", "3.0 (quilt)\n"),
            ("debian/source/options", 'compression = "gzip"\n')])
        source.add(["debian/source/format", "debian/source/options"])
        source.commit("set source format")
        return source

    def test_cmd_dailydeb_missing_orig_tarball(self):
        self.make_simple_quilt_package()
        with open("test.recipe", "w") as recipe:
            recipe.write(
                "# git-build-recipe format 0.3 deb-version 1-1\nsource\n")
        out, err = self.run_recipe("test.recipe working", retcode=2)
        self.assertIn(
            'error: Unable to find the upstream source.  '
            'Import it as tag upstream/1 or build with '
            '--allow-fallback-to-native.\n', err)

    @skipIfFakeroot()
    def test_cmd_dailydeb_with_orig_tarball(self):
        source = self.make_simple_package("source")
        self.make_upstream_version(
            source, "package", "0.1", [("file", "content\n")])
        with open("test.recipe", "w") as recipe:
            recipe.write(
                "# git-build-recipe format 0.3 deb-version 0.1-1\nsource\n")
        out, err = self.run_recipe("test.recipe working")
        self.assertThat("working/package_0.1.orig.tar.gz", PathExists())
        self.assertThat("working/package_0.1-1.diff.gz", PathExists())

    @skipIfFakeroot()
    @requirePristineTar()
    def test_cmd_dailydeb_with_pristine_orig_gz_tarball(self):
        source = self.make_simple_package("source")
        pristine_tar_sha1 = self.make_upstream_version(
            source, "package", "0.1", [("file", "content\n")],
            pristine_tar_format="gz")
        with open("test.recipe", "w") as recipe:
            recipe.write(
                "# git-build-recipe format 0.3 deb-version 0.1-1\nsource\n")
        out, err = self.run_recipe("test.recipe working")
        self.assertThat("working/package_0.1.orig.tar.gz", PathExists())
        self.assertThat("working/package_0.1-1.diff.gz", PathExists())
        self.assertEqual(
            sha1_file_by_name("working/package_0.1.orig.tar.gz"),
            pristine_tar_sha1)

    @skipIfFakeroot()
    @requirePristineTar()
    def test_cmd_dailydeb_with_pristine_orig_bz2_tarball(self):
        source = self.make_simple_quilt_package()
        pristine_tar_sha1 = self.make_upstream_version(
            source, "package", "0.1",
            [("file", "content\n"), ("a", "contents of a\n")],
            pristine_tar_format="bz2")
        with open("test.recipe", "w") as recipe:
            recipe.write(
                "# git-build-recipe format 0.3 deb-version 0.1-1\nsource\n")
        source.build_tree([("upstream/")])
        source.build_tree_contents([
            ("upstream/file", "content\n"),
            ("upstream/a", "contents of a\n")])
        source.add(["upstream/file", "upstream/a"])
        out, err = self.run_recipe("test.recipe working")
        self.assertThat("working/package_0.1.orig.tar.bz2", PathExists())
        self.assertThat("working/package_0.1-1.debian.tar.gz", PathExists())
        self.assertEqual(
            sha1_file_by_name("working/package_0.1.orig.tar.bz2"),
            pristine_tar_sha1)

    @requirePristineTar()
    def test_cmd_dailydeb_allow_fallback_to_native_with_orig_tarball(self):
        source = self.make_simple_quilt_package()
        pristine_tar_sha1 = self.make_upstream_version(
            source, "package", "0.1",
            [("file", "content\n"), ("a", "contents of a\n")],
            pristine_tar_format="bz2")
        with open("test.recipe", "w") as recipe:
            recipe.write(
                "# git-build-recipe format 0.3 deb-version 0.1-1\nsource\n")
        source.build_tree([("upstream/")])
        source.build_tree_contents([
            ("upstream/file", "content\n"),
            ("upstream/a", "contents of a\n")])
        source.add(["upstream/file", "upstream/a"])
        out, err = self.run_recipe(
            "--allow-fallback-to-native test.recipe working")
        self.assertThat("working/package_0.1.orig.tar.bz2", PathExists())
        self.assertThat("working/package_0.1-1.debian.tar.gz", PathExists())
        self.assertThat(
            "source/debian/source/format", FileContains("3.0 (quilt)\n"))
        self.assertEqual(
            sha1_file_by_name("working/package_0.1.orig.tar.bz2"),
            pristine_tar_sha1)

    @skipIfFakeroot()
    def test_cmd_dailydeb_force_native(self):
        self.make_simple_quilt_package()
        with open("test.recipe", "w") as recipe:
            recipe.write(
                "# git-build-recipe format 0.3 deb-version 1\nsource\n")
        out, err = self.run_recipe(
            "--allow-fallback-to-native test.recipe working")
        self.assertThat(
            "working/test/debian/source/format",
            FileContains("3.0 (native)\n"))

    @skipIfFakeroot()
    def test_cmd_dailydeb_force_native_empty_series(self):
        source = self.make_simple_quilt_package()
        source.build_tree(['debian/patches/'])
        source.build_tree_contents([("debian/patches/series", "\n")])
        source.add(["debian/patches/series"])
        source.commit("add patches")
        with open("test.recipe", "w") as recipe:
            recipe.write(
                "# git-build-recipe format 0.3 deb-version 1\nsource\n")
        out, err = self.run_recipe(
            "--allow-fallback-to-native test.recipe working")
        self.assertThat(
            "working/test/debian/source/format",
            FileContains("3.0 (native)\n"))
        self.assertThat("working/test/debian/patches", Not(PathExists()))

    @skipIfFakeroot()
    def test_cmd_dailydeb_force_native_apply_quilt(self):
        source = self.make_simple_quilt_package()
        source.build_tree(["debian/patches/"])
        patch = dedent("""\
            diff -ur a/thefile b/thefile
            --- a/thefile	2010-12-05 20:14:22.000000000 +0100
            +++ b/thefile	2010-12-05 20:14:26.000000000 +0100
            @@ -1 +1 @@
            -old-contents
            +new-contents
            """)
        source.build_tree_contents([
            ("thefile", "old-contents\n"),
            ("debian/patches/series", "01_foo.patch"),
            ("debian/patches/01_foo.patch", patch)])
        source.add([
            "thefile", "debian/patches/series",
            "debian/patches/01_foo.patch"])
        source.commit("add patch")
        with open("test.recipe", "w") as recipe:
            recipe.write(
                "# git-build-recipe format 0.3 deb-version 1\nsource\n")
        out, err = self.run_recipe(
            "--allow-fallback-to-native test.recipe working")
        self.assertThat(
            "working/test/debian/source/format",
            FileContains("3.0 (native)\n"))
        self.assertThat("working/test/thefile", FileContains("new-contents\n"))
        self.assertThat("working/test/debian/patches", Not(PathExists()))

    def test_cmd_dailydeb_force_native_apply_quilt_failure(self):
        source = self.make_simple_quilt_package()
        source.build_tree(["debian/patches/"])
        patch = dedent("""\
            diff -ur a/thefile b/thefile
            --- a/thefile	2010-12-05 20:14:22.000000000 +0100
            +++ b/thefile	2010-12-05 20:14:26.000000000 +0100
            @@ -1 +1 @@
            -old-contents
            +new-contents
            """)
        source.build_tree_contents([
            ("thefile", "contents\n"),
            ("debian/patches/series", "01_foo.patch"),
            ("debian/patches/01_foo.patch", patch)])
        source.add([
            "thefile", "debian/patches/series",
            "debian/patches/01_foo.patch"])
        source.commit("add patch")
        with open("test.recipe", "w") as recipe:
            recipe.write(
                "# git-build-recipe format 0.3 deb-version 1-1\nsource\n")
        out, err = self.run_recipe(
            "--allow-fallback-to-native test.recipe working", retcode=1)
        self.assertIn("Failed to apply quilt patches\n", err)

    def test_unknown_source_format(self):
        source = self.make_simple_package("source")
        source.build_tree(["debian/source/"])
        source.build_tree_contents([("debian/source/format", "2.0\n")])
        source.add(["debian/source/format"])
        source.commit("set source format")
        with open("test.recipe", "w") as recipe:
            recipe.write(
                "# git-build-recipe format 0.3 deb-version 1-1\nsource\n")
        out, err = self.run_recipe(
            "--allow-fallback-to-native test.recipe working", retcode=1)
        self.assertIn("Unknown source format 2.0\n", err)



# Copyright 2015 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it 
# under the terms of the GNU General Public License version 3, as published 
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but 
# WITHOUT ANY WARRANTY; without even the implied warranties of 
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR 
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along 
# with this program.  If not, see <http://www.gnu.org/licenses/>.

import datetime
import subprocess
import textwrap

from debian import changelog

from gitbuildrecipe.deb_version import (
    DebUpstreamBaseVariable,
    DebUpstreamVariable,
    DebVersionVariable,
    NotFullyExpanded,
    SubstitutionUnavailable,
    check_expanded_deb_version,
    version_extract_base,
    substitute_branch_vars,
    substitute_time,
    )
from gitbuildrecipe.recipe import (
    BaseRecipeBranch,
    build_tree,
    RecipeBranch,
    resolve_revisions,
    )
from gitbuildrecipe.tests import (
    GitRepository,
    GitTestCase,
    )


class ResolveRevisionsTests(GitTestCase):

    def setUp(self):
        super().setUp()
        self.use_temp_dir()

    def test_unchanged(self):
        source = GitRepository("source")
        commit = source.commit("one", date="2015-01-01 12:00:00 UTC")
        branch1 = BaseRecipeBranch("source", "{revdate}", 0.4, revspec="HEAD")
        branch2 = BaseRecipeBranch("source", "{revdate}", 0.4, revspec=commit)
        self.assertFalse(resolve_revisions(
            branch1, if_changed_from=branch2,
            substitute_branch_vars=substitute_branch_vars))
        self.assertEqual("source", branch1.url)
        self.assertEqual(commit, branch1.commit)
        self.assertEqual("HEAD", branch1.revspec)
        self.assertEqual("20150101", branch1.deb_version)

    def test_unchanged_not_explicit(self):
        source = GitRepository("source")
        commit = source.commit("one", date="2015-01-01 12:00:00 UTC")
        branch1 = BaseRecipeBranch("source", "{revdate}", 0.4)
        branch2 = BaseRecipeBranch("source", "{revdate}", 0.4, revspec=commit)
        self.assertFalse(resolve_revisions(
            branch1, if_changed_from=branch2,
            substitute_branch_vars=substitute_branch_vars))
        self.assertEqual("source", branch1.url)
        self.assertEqual(commit, branch1.commit)
        self.assertIsNone(branch1.revspec)
        self.assertEqual("20150101", branch1.deb_version)

    def test_unchanged_multilevel(self):
        source = GitRepository("source")
        commit = source.commit("one", date="2015-01-01 12:00:00 UTC")
        branch1 = BaseRecipeBranch("source", "{revdate}", 0.4)
        branch2 = RecipeBranch("nested1", "source")
        branch3 = RecipeBranch("nested2", "source")
        branch2.nest_branch("bar", branch3)
        branch1.nest_branch("foo", branch2)
        branch4 = BaseRecipeBranch("source", "{revdate}", 0.4, revspec=commit)
        branch5 = RecipeBranch("nested1", "source", revspec=commit)
        branch6 = RecipeBranch("nested2", "source", revspec=commit)
        branch5.nest_branch("bar", branch6)
        branch4.nest_branch("foo", branch5)
        self.assertFalse(resolve_revisions(
            branch1, if_changed_from=branch4,
            substitute_branch_vars=substitute_branch_vars))
        self.assertEqual("source", branch1.url)
        self.assertEqual(commit, branch1.commit)
        self.assertIsNone(branch1.revspec)
        self.assertEqual("20150101", branch1.deb_version)

    def test_changed(self):
        source = GitRepository("source")
        commit = source.commit("one", date="2015-01-01 12:00:00 UTC")
        branch1 = BaseRecipeBranch("source", "{revdate}", 0.4, revspec="HEAD")
        branch2 = BaseRecipeBranch("source", "{revdate}", 0.4, revspec="foo")
        self.assertTrue(resolve_revisions(
            branch1, if_changed_from=branch2,
            substitute_branch_vars=substitute_branch_vars))
        self.assertEqual("source", branch1.url)
        self.assertEqual(commit, branch1.commit)
        self.assertEqual("HEAD", branch1.revspec)
        self.assertEqual("20150101", branch1.deb_version)

    def test_changed_shape(self):
        source = GitRepository("source")
        commit = source.commit("one", date="2015-01-01 12:00:00 UTC")
        branch1 = BaseRecipeBranch("source", "{revdate}", 0.4, revspec="HEAD")
        branch2 = BaseRecipeBranch("source", "{revdate}", 0.4, revspec=commit)
        branch3 = RecipeBranch("nested", "source")
        branch1.nest_branch("foo", branch3)
        self.assertTrue(resolve_revisions(
            branch1, if_changed_from=branch2,
            substitute_branch_vars=substitute_branch_vars))
        self.assertEqual("source", branch1.url)
        self.assertEqual(commit, branch1.commit)
        self.assertEqual("HEAD", branch1.revspec)
        self.assertEqual("20150101", branch1.deb_version)

    def test_changed_command(self):
        source = GitRepository("source")
        source.commit("one")
        branch1 = BaseRecipeBranch("source", "{revdate}", 0.4)
        branch2 = BaseRecipeBranch("source", "{revdate}", 0.4)
        branch1.run_command("touch test1")
        branch2.run_command("touch test2")
        self.assertTrue(resolve_revisions(
            branch1, if_changed_from=branch2,
            substitute_branch_vars=substitute_branch_vars))
        self.assertEqual("source", branch1.url)

    def test_unchanged_command(self):
        source = GitRepository("source")
        source.commit("one")
        branch1 = BaseRecipeBranch("source", "{revdate}", 0.4)
        branch2 = BaseRecipeBranch("source", "{revdate}", 0.4)
        branch1.run_command("touch test1")
        branch2.run_command("touch test1")
        self.assertFalse(resolve_revisions(
            branch1, if_changed_from=branch2,
            substitute_branch_vars=substitute_branch_vars))
        self.assertEqual("source", branch1.url)

    def test_substitute(self):
        source = GitRepository("source")
        commit = source.commit("one", date="2015-01-01 12:00:00 UTC")
        source.commit("two", date="2015-02-01 12:00:00 UTC")
        branch1 = BaseRecipeBranch(
            "source", "{revdate}-{revdate:packaging}", 0.4, revspec=commit)
        branch2 = RecipeBranch("packaging", "source")
        branch1.nest_branch("debian", branch2)
        self.assertTrue(resolve_revisions(
            branch1, substitute_branch_vars=substitute_branch_vars))
        self.assertEqual("source", branch1.url)
        self.assertEqual(commit, branch1.commit)
        self.assertEqual(commit, branch1.revspec)
        self.assertEqual("20150101-20150201", branch1.deb_version)

    def test_substitute_supports_debupstream(self):
        # resolve_revisions should leave debupstream parameters alone and not
        # complain.
        source = GitRepository("source")
        source.commit("one", date="2015-01-01 12:00:00 UTC")
        source.commit("two", date="2015-01-02 12:00:00 UTC")
        branch1 = BaseRecipeBranch("source", "{debupstream}-{revdate}", 0.4)
        resolve_revisions(
            branch1, substitute_branch_vars=substitute_branch_vars)
        self.assertEqual("{debupstream}-20150102", branch1.deb_version)

    def test_substitute_not_fully_expanded(self):
        source = GitRepository("source")
        source.commit("one")
        source.commit("two")
        branch1 = BaseRecipeBranch("source", "{revdate:packaging}", 0.4)
        resolve_revisions(
            branch1, substitute_branch_vars=substitute_branch_vars)
        self.assertRaises(
            NotFullyExpanded, check_expanded_deb_version, branch1)

    def test_substitute_git(self):
        source = GitRepository("source")
        commit = source.commit("one")
        branch1 = BaseRecipeBranch("source", "foo-{git-commit}", 0.4)
        resolve_revisions(
            branch1, substitute_branch_vars=substitute_branch_vars)
        self.assertEqual("foo-%s" % commit[:7], branch1.deb_version)

    def test_latest_tag(self):
        source = GitRepository("source")
        commit = source.commit("one")
        source.tag("millbank", commit)
        source.commit("two")
        branch1 = BaseRecipeBranch("source", "foo-{latest-tag}", 0.4)
        resolve_revisions(branch1,
            substitute_branch_vars=substitute_branch_vars)
        self.assertEqual("foo-millbank", branch1.deb_version)

    def test_latest_tag_no_tag(self):
        source = GitRepository("source")
        source.commit("one")
        source.commit("two")
        branch1 = BaseRecipeBranch("source", "foo-{latest-tag}", 0.4)
        exc = self.assertRaises(
            subprocess.CalledProcessError, resolve_revisions,
            branch1, substitute_branch_vars=substitute_branch_vars)
        self.assertIn("No names found, cannot describe anything.", exc.output)

    def test_substitute_revdate(self):
        source = GitRepository("source")
        source.commit("one")
        source.commit("two", date="2011-06-10 12:23:48 UTC")
        branch1 = BaseRecipeBranch("source", "foo-{revdate}", 0.4)
        resolve_revisions(
            branch1, substitute_branch_vars=substitute_branch_vars)
        self.assertEqual("foo-20110610", branch1.deb_version)

    def test_substitute_revtime(self):
        source = GitRepository("source")
        source.commit("one")
        source.commit("two", date="2011-06-10 12:23:48 UTC")
        branch1 = BaseRecipeBranch("source", "foo-{revtime}", 0.4)
        resolve_revisions(
            branch1, substitute_branch_vars=substitute_branch_vars)
        self.assertEqual("foo-201106101223", branch1.deb_version)

    def test_substitute_revno(self):
        source = GitRepository("source")
        source.commit("one")
        source._git_call("checkout", "-q", "-b", "branch")
        source.build_tree(["a"])
        source.add(["a"])
        source.commit("two")
        source.commit("three")
        branch1 = BaseRecipeBranch("source", "foo-{revno}", 0.1)
        resolve_revisions(
            branch1, substitute_branch_vars=substitute_branch_vars)
        self.assertEqual("foo-3", branch1.deb_version)
        source._git_call("checkout", "-q", "master")
        source._git_call(
            "merge", "-q", "--no-ff", "--commit", "-m", "merge branch",
            "branch")
        branch2 = BaseRecipeBranch("source", "foo-{revno}", 0.1)
        resolve_revisions(
            branch2, substitute_branch_vars=substitute_branch_vars)
        self.assertEqual("foo-2", branch2.deb_version)

    def test_uses_source_commit(self):
        # The original commit is used for variable substitution, even if
        # the base branch was committed to by instructions such as merge.
        source = GitRepository("source")
        source.commit("one")
        source._git_call("checkout", "-q", "-b", "branch")
        source.build_tree(["a"])
        source.add(["a"])
        source.commit("two")
        source.commit("three")
        branch1 = BaseRecipeBranch(
            "source", "foo-{revno}+{revno:branch}", 0.1, revspec="master")
        branch2 = RecipeBranch("branch", "source", revspec="branch")
        branch1.merge_branch(branch2)
        build_tree(branch1, "target")
        resolve_revisions(
            branch1, substitute_branch_vars=substitute_branch_vars)
        self.assertEqual("foo-1+3", branch1.deb_version)


class DebUpstreamVariableTests(GitTestCase):

    def write_changelog(self, version):
        contents = textwrap.dedent("""
            package (%s) experimental; urgency=low

              * Initial release. (Closes: #XXXXXX)

             -- Jelmer Vernooij <jelmer@debian.org>  Thu, 19 May 2011 10:07:41 +0100
            """ % version)[1:]
        return changelog.Changelog(file=contents)

    def test_empty_changelog(self):
        var = DebUpstreamVariable.from_changelog(None, changelog.Changelog())
        self.assertRaises(SubstitutionUnavailable, var.get)

    def test_version(self):
        var = DebUpstreamVariable.from_changelog(
            None, self.write_changelog("2.3"))
        self.assertEqual("2.3", var.get())

    def test_epoch(self):
        # The epoch is (currently) ignored by {debupstream}.
        var = DebUpstreamVariable.from_changelog(
            None, self.write_changelog("2:2.3"))
        self.assertEqual("2.3", var.get())

    def test_base_without_snapshot(self):
        var = DebUpstreamBaseVariable.from_changelog(
            None, self.write_changelog("2.4"))
        self.assertEqual("2.4+", var.get())

    def test_base_with_svn_snapshot(self):
        var = DebUpstreamBaseVariable.from_changelog(
            None, self.write_changelog("2.4~svn4"))
        self.assertEqual("2.4~", var.get())

    def test_base_with_bzr_snapshot(self):
        var = DebUpstreamBaseVariable.from_changelog(
            None, self.write_changelog("2.4+bzr343"))
        self.assertEqual("2.4+", var.get())


class VersionExtractBaseTests(GitTestCase):

    def test_simple_extract(self):
        self.assertEqual("2.4", version_extract_base("2.4"))
        self.assertEqual("2.4+foobar", version_extract_base("2.4+foobar"))

    def test_with_bzr(self):
        self.assertEqual("2.4+", version_extract_base("2.4+bzr32"))
        self.assertEqual("2.4~", version_extract_base("2.4~bzr32"))

    def test_with_git(self):
        self.assertEqual("2.4+", version_extract_base("2.4+git20101010"))
        self.assertEqual("2.4~", version_extract_base("2.4~gitaabbccdd"))

    def test_with_svn(self):
        self.assertEqual("2.4+", version_extract_base("2.4+svn45"))
        self.assertEqual("2.4~", version_extract_base("2.4~svn45"))

    def test_with_dfsg(self):
        self.assertEqual("2.4+", version_extract_base("2.4+bzr32+dfsg1"))
        self.assertEqual("2.4~", version_extract_base("2.4~bzr32+dfsg.1"))
        self.assertEqual("2.4~", version_extract_base("2.4~bzr32.dfsg.1"))
        self.assertEqual("2.4~", version_extract_base("2.4~bzr32dfsg.1"))
        self.assertEqual("1.6~", version_extract_base("1.6~git20120320.dfsg.1"))


class DebVersionVariableTests(GitTestCase):

    def write_changelog(self, version):
        contents = textwrap.dedent("""
            package (%s) experimental; urgency=low

              * Initial release. (Closes: #XXXXXX)

             -- Jelmer Vernooij <jelmer@debian.org>  Thu, 19 May 2011 10:07:41 +0100
            """ % version)[1:]
        return changelog.Changelog(file=contents)

    def test_empty_changelog(self):
        var = DebVersionVariable.from_changelog(None, changelog.Changelog())
        self.assertRaises(SubstitutionUnavailable, var.get)

    def test_simple(self):
        var = DebVersionVariable.from_changelog(
            None, self.write_changelog("2.3-1"))
        self.assertEqual("2.3-1", var.get())

    def test_epoch(self):
        var = DebVersionVariable.from_changelog(
            None, self.write_changelog("4:2.3-1"))
        self.assertEqual("4:2.3-1", var.get())


class RecipeBranchTests(GitTestCase):

    def setUp(self):
        super().setUp()
        self.use_temp_dir()

    def test_substitute_time(self):
        time = datetime.datetime.utcfromtimestamp(1)
        base_branch = BaseRecipeBranch("base_url", "1-{time}", 0.2)
        substitute_time(base_branch, time)
        self.assertEqual("1-197001010000", base_branch.deb_version)
        substitute_time(base_branch, time)
        self.assertEqual("1-197001010000", base_branch.deb_version)

    def test_substitute_date(self):
        time = datetime.datetime.utcfromtimestamp(1)
        base_branch = BaseRecipeBranch("base_url", "1-{date}", 0.2)
        substitute_time(base_branch, time)
        self.assertEqual("1-19700101", base_branch.deb_version)
        substitute_time(base_branch, time)
        self.assertEqual("1-19700101", base_branch.deb_version)

    def test_substitute_branch_vars(self):
        wt = GitRepository("base_url")
        commit1 = wt.commit("acommit", date="2015-01-01 12:00:00 UTC")
        commit2 = wt.commit("bcommit", date="2015-01-02 12:00:00 UTC")
        base_branch = BaseRecipeBranch("base_url", "1", 0.2, revspec=commit1)
        substitute_branch_vars(base_branch, base_branch)
        self.assertEqual("1", base_branch.deb_version)
        substitute_branch_vars(base_branch, base_branch)
        self.assertEqual("1", base_branch.deb_version)
        base_branch = BaseRecipeBranch(
            "base_url", "{revdate}", 0.4, revspec=commit1)
        base_branch.resolve_commit()
        substitute_branch_vars(base_branch, base_branch)
        self.assertEqual("20150101", base_branch.deb_version)
        base_branch = BaseRecipeBranch(
            "base_url", "{revdate}", 0.4, revspec=commit1)
        base_branch.resolve_commit()
        child_branch = RecipeBranch("foo", "base_url", revspec=commit2)
        child_branch.resolve_commit()
        substitute_branch_vars(base_branch, child_branch)
        self.assertEqual("{revdate}", base_branch.deb_version)
        substitute_branch_vars(base_branch, child_branch)
        self.assertEqual("{revdate}", base_branch.deb_version)
        base_branch = BaseRecipeBranch(
            "base_url", "{revdate:foo}", 0.4, revspec=commit1)
        base_branch.resolve_commit()
        substitute_branch_vars(base_branch, child_branch)
        self.assertEqual("20150102", base_branch.deb_version)

    def test_substitute_branch_vars_debupstream(self):
        wt = GitRepository("base_url")
        commit1 = wt.commit("acommit")
        cl_contents = ("package (0.1-1) unstable; urgency=low\n  * foo\n"
                    " -- maint <maint@maint.org>  Tue, 04 Aug 2009 "
                    "10:03:10 +0100\n")
        wt.build_tree_contents(
            [("debian/", ), ('debian/changelog', cl_contents)])
        wt.add(['debian/changelog'])
        commit2 = wt.commit("with changelog")
        base_branch = BaseRecipeBranch(
            "base_url", "{debupstream}", 0.4, revspec=commit1)
        # No changelog file, so no substitution
        substitute_branch_vars(base_branch, base_branch)
        self.assertEqual("{debupstream}", base_branch.deb_version)
        base_branch = BaseRecipeBranch(
            "base_url", "{debupstream}", 0.4, revspec=commit2)
        substitute_branch_vars(base_branch, base_branch)
        self.assertEqual("0.1", base_branch.deb_version)
        base_branch = BaseRecipeBranch(
            "base_url", "{debupstream:tehname}", 0.4)
        child_branch = RecipeBranch("tehname", "base_url", revspec=commit2)
        substitute_branch_vars(base_branch, child_branch)
        self.assertEqual("0.1", base_branch.deb_version)

    def test_substitute_branch_vars_debupstream_pre_0_4(self):
        wt = GitRepository("base_url")
        cl_contents = ("package (0.1-1) unstable; urgency=low\n  * foo\n"
                    " -- maint <maint@maint.org>  Tue, 04 Aug 2009 "
                    "10:03:10 +0100\n")
        wt.build_tree_contents(
            [("debian/", ), ('debian/changelog', cl_contents)])
        wt.add(['debian/changelog'])
        commit = wt.commit("with changelog")
        # In recipe format < 0.4 {debupstream} gets replaced from the
        # resulting tree, not from the branch vars.
        base_branch = BaseRecipeBranch(
            "base_url", "{debupstream}", 0.2, revspec=commit)
        substitute_branch_vars(base_branch, base_branch)
        self.assertEqual("{debupstream}", base_branch.deb_version)



#! /usr/bin/env python3

# Copyright 2015 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it 
# under the terms of the GNU General Public License version 3, as published 
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but 
# WITHOUT ANY WARRANTY; without even the implied warranties of 
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR 
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along 
# with this program.  If not, see <http://www.gnu.org/licenses/>.

from setuptools import setup

requires = [
    'python-dateutil',
    'python-debian',
    ]
test_requires = [
    'fixtures',
    'testtools',
    ]

setup(
    name='git-build-recipe',
    version='0.1',
    packages=['gitbuildrecipe', 'gitbuildrecipe.tests'],
    entry_points={
        'console_scripts': ['git-build-recipe = gitbuildrecipe.main:main'],
        },
    zip_safe=False,
    maintainer='Colin Watson',
    maintainer_email='cjwatson@ubuntu.com',
    description='git-build-recipe',
    license='GPL v3',
    url='https://launchpad.net/git-build-recipe',
    download_url='https://launchpad.net/git-build-recipe/+download',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3',
        'Operating System :: OS Independent',
        'Programming Language :: Python 3',
        ],
    install_requires=requires,
    tests_require=test_requires,
    extras_require={"test": test_requires},
    test_suite='gitbuildrecipe.tests',
    )



# Copyright 2015 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it 
# under the terms of the GNU General Public License version 3, as published 
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but 
# WITHOUT ANY WARRANTY; without even the implied warranties of 
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR 
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along 
# with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Handle recipe parsing and substitution variables."""

import datetime
from functools import lru_cache
import logging
import os
import subprocess
import sys
from urllib.parse import urlparse

import dateutil.parser
from debian import debian_support


MERGE_INSTRUCTION = "merge"
NEST_PART_INSTRUCTION = "nest-part"
NEST_INSTRUCTION = "nest"
RUN_INSTRUCTION = "run"
USAGE = {
    MERGE_INSTRUCTION: 'merge NAME BRANCH [REVISION]',
    NEST_INSTRUCTION: 'nest NAME BRANCH TARGET-DIR [REVISION]',
    NEST_PART_INSTRUCTION:
        'nest-part NAME BRANCH SUBDIR [TARGET-DIR [REVISION]]',
    RUN_INSTRUCTION: 'run COMMAND',
    }

SAFE_INSTRUCTIONS = [
    MERGE_INSTRUCTION, NEST_PART_INSTRUCTION, NEST_INSTRUCTION]


class FormattedError(Exception):

    def __init__(self, **kwargs):
        super().__init__()
        for key, value in kwargs.items():
            setattr(self, key, value)

    def __str__(self):
        return self._fmt % dict(self.__dict__)

    def __repr__(self):
        return "%s(%s)" % (self.__class__.__name__, str(self))

    def __eq__(self, other):
        if self.__class__ is not other.__class__:
            return NotImplemented
        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        if self.__class__ is not other.__class__:
            return NotImplemented
        return self.__dict__ != other.__dict__

    __hash__ = Exception.__hash__


class SubstitutionUnavailable(FormattedError):

    _fmt = "Substitution for %(name)s not available: %(reason)s"

    def __init__(self, name, reason):
        super().__init__(name=name, reason=reason)


class SubstitutionVariable:
    """A substitution variable for a version string."""

    def replace(self, value):
        """Replace name with value."""
        raise NotImplementedError(self.replace)

    @classmethod
    def available_in(cls, format):
        """Check if this variable is available in a particular format."""
        raise NotImplementedError(cls.available_in)


class SimpleSubstitutionVariable(SubstitutionVariable):

    name = None

    minimum_format = None

    def replace(self, value):
        if not self.name in value:
            return value
        return value.replace(self.name, self.get())

    def get(self):
        raise NotImplementedError(self.value)

    @classmethod
    def available_in(self, format):
        return format >= self.minimum_format


class BranchSubstitutionVariable(SimpleSubstitutionVariable):

    basename = None

    def __init__(self, branch_name=None):
        super().__init__()
        self.branch_name = branch_name

    @classmethod
    def determine_name(cls, branch_name):
        if branch_name is None:
            return "{%s}" % cls.basename
        else:
            return "{%s:%s}" % (cls.basename, branch_name)

    @property
    def name(self):
        return self.determine_name(self.branch_name)


class TimeVariable(SimpleSubstitutionVariable):

    name = "{time}"

    minimum_format = 0.1

    def __init__(self, time):
        self._time = time

    def get(self):
        return self._time.strftime("%Y%m%d%H%M")


class DateVariable(SimpleSubstitutionVariable):

    name = "{date}"

    minimum_format = 0.1

    def __init__(self, time):
        self._time = time

    def get(self):
        return self._time.strftime("%Y%m%d")


class RevisionVariable(BranchSubstitutionVariable):

    minimum_format = 0.1

    def __init__(self, branch):
        super().__init__(branch.name)
        self.branch = branch
        self.commit = branch.commit


class RevnoVariable(RevisionVariable):
    """A {revno} substitution variable.

    This is a distinctly un-git-like bzr-ism, and is therefore a hack.  Most
    users should use {revtime} or {revdate} instead, which is monotonically
    increasing, or perhaps also {git-commit} to ensure uniqueness (but note
    that commit hashes do not increase monotonically, so are not normally
    suitable in version strings).  However, this can result in very long
    version strings, especially if there are multiple branches involved; so
    this may for instance be useful as {revno:packaging} to encode the
    length of the packaging branch.

    This produces an approximation of a bzr revno by taking the length of
    the left-hand history.
    """

    basename = "revno"

    minimum_format = 0.1

    def get(self):
        return self.branch.git_output(
            "rev-list", "--first-parent", "--count", self.commit).rstrip("\n")


class RevtimeVariable(RevisionVariable):

    basename = "revtime"

    minimum_format = 0.4

    def get(self):
        dt = dateutil.parser.parse(self.branch.git_output(
            "log", "-1", "--date=iso", "--format=format:%cd", self.commit))
        return dt.astimezone(datetime.timezone.utc).strftime("%Y%m%d%H%M")


class RevdateVariable(RevisionVariable):

    basename = "revdate"

    minimum_format = 0.4

    def get(self):
        dt = dateutil.parser.parse(self.branch.git_output(
            "log", "-1", "--date=iso", "--format=format:%cd", self.commit))
        return dt.astimezone(datetime.timezone.utc).strftime("%Y%m%d")


class GitCommitVariable(RevisionVariable):

    basename = "git-commit"

    minimum_format = 0.4

    def get(self):
        return self.branch.git_output(
            "rev-parse", "--short", self.commit).rstrip("\n")


class LatestTagVariable(RevisionVariable):

    basename = "latest-tag"

    minimum_format = 0.4

    def get(self):
        # Capture stderr so that exceptions are useful.  We know that git
        # describe does not write to stderr on success.
        return self.branch.git_output(
            "describe", "--tags", "--abbrev=0", self.commit,
            stderr=subprocess.STDOUT).rstrip("\n")


branch_vars = [
    GitCommitVariable,
    LatestTagVariable,
    RevdateVariable,
    RevtimeVariable,
    ]
simple_vars = [
    TimeVariable,
    DateVariable,
    ]


class TargetAlreadyExists(FormattedError):

    _fmt = "%(target)s already exists, but has no .git directory."

    def __init__(self, target):
        super().__init__(target=target)


class GitCommandFailed(FormattedError):

    _fmt = "git %(command)s failed:\n%(output)s"

    def __init__(self, output):
        super().__init__(command=self._cmd, output=output)


class FetchFailed(GitCommandFailed):

    _cmd = "fetch"


class CloneFailed(GitCommandFailed):

    _cmd = "clone"


class CheckoutFailed(GitCommandFailed):

    _cmd = "checkout"


class MergeFailed(GitCommandFailed):

    _cmd = "merge"


# Some default configuration to make writing recipes more pleasant.
insteadof = {
    "lp:": "https://git.launchpad.net/"
    }


def pull_or_clone(base_branch, target_path):
    """Make target_path match base_branch, either by pulling or by cloning."""
    base_branch.path = target_path
    if os.path.exists(target_path) and not os.listdir(target_path):
        os.rmdir(target_path)
    if not os.path.exists(target_path):
        os.makedirs(target_path)
        base_branch.git_call("init")
        for short, base in insteadof.items():
            base_branch.git_call("config", "url.%s.insteadOf" % base, short)
    elif not os.path.exists(os.path.join(target_path, ".git")):
        raise TargetAlreadyExists(target_path)
    try:
        fetch_branches(base_branch)
    except subprocess.CalledProcessError as e:
        raise FetchFailed(e.output)
    try:
        base_branch.resolve_commit()
        # Check out the commit hash directly to detach HEAD and ensure
        # commits (eg. by a merge instruction) don't affect substitution
        # variables.
        # If someone sneakily names a branch after a commit then all
        # hell will break loose (really, git? there's no way to force
        # checkout to interpret its argument as a commit?), but that's
        # their problem.
        base_branch.git_call("checkout", base_branch.commit)
    except subprocess.CalledProcessError as e:
        raise CheckoutFailed(e.output)


def fetch_branches(child_branch):
    url = child_branch.url
    parsed_url = urlparse(url)
    if not parsed_url.scheme and not parsed_url.path.startswith("/"):
        url = os.path.abspath(url)
    # Fetch all remote branches, the remote HEAD, and (implicitly) any
    # tags that reference commits in those refs. Tags that aren't on a
    # branch won't be fetched.
    child_branch.git_call(
        "fetch", url,
        "HEAD:refs/remotes/%s/HEAD" % child_branch.remote_name,
        "refs/heads/*:refs/remotes/%s/*" % child_branch.remote_name)


@lru_cache(maxsize=1)
def _git_version():
    raw_git_version = subprocess.check_output(
        ["dpkg-query", "-W", "-f", "${Version}", "git"],
        universal_newlines=True)
    return debian_support.Version(raw_git_version)


def merge_branch(child_branch, target_path):
    """Merge the branch specified by `child_branch`.

    :param child_branch: A `RecipeBranch` identifying the branch and commit
        to merge from.
    :param target_path: The tree to merge into.
    """
    child_branch.path = target_path
    fetch_branches(child_branch)
    try:
        child_branch.resolve_commit()
        cmd = ["merge", "--commit"]
        if _git_version() >= "1:2.9.0":
            cmd.append("--allow-unrelated-histories")
        cmd.extend(["-m", "Merge %s" % child_branch.get_revspec()])
        cmd.append(child_branch.commit)
        child_branch.git_call(*cmd)
    except subprocess.CalledProcessError as e:
        raise MergeFailed(e.output)


def nest_part_branch(child_branch, target_path, subpath, target_subdir=None):
    """Nest the branch subdirectory specified by `child_branch`.

    :param child_branch: A `RecipeBranch` identifying the branch and commit
        to merge from.
    :param target_path: The tree to merge into.
    :param subpath: Only merge files from `child_branch` that are from this
        path.  For example, subpath="/debian" will only merge changes from
        that directory.
    :param target_subdir: Optional directory in `target_path` to merge that
        subpath into.  Defaults to the basename of `subpath`.
    """
    subpath = subpath.lstrip("/")
    if target_subdir is None:
        target_subdir = os.path.basename(subpath)
    # XXX should handle updating as well
    assert not os.path.exists(target_subdir)
    child_branch.path = target_path
    fetch_branches(child_branch)
    child_branch.resolve_commit()
    child_branch.git_call(
        "read-tree", "--prefix", target_subdir, "-u",
        child_branch.commit + ":" + subpath)


def _build_inner_tree(base_branch, target_path):
    revision_of = ""
    if base_branch.revspec is not None:
        revision_of = "revision '%s' of " % base_branch.revspec
    logging.info(
        "Retrieving %s'%s' to put at '%s'." %
        (revision_of, base_branch.url, target_path))
    pull_or_clone(base_branch, target_path)
    base_branch.resolve_commit()
    for instruction in base_branch.child_branches:
        instruction.apply(target_path)


def _resolve_revisions_recurse(new_branch, substitute_branch_vars,
                               if_changed_from=None):
    changed = False
    if substitute_branch_vars is not None:
        # XXX need to make sure new_branch has been fetched
        new_branch.resolve_commit()
        substitute_branch_vars(new_branch)
    if (if_changed_from is not None and
        (new_branch.revspec is not None or
         if_changed_from.revspec is not None)):
        if if_changed_from.revspec is not None:
            # XXX resolve revspec
            changed_commit = if_changed_from.revspec
        else:
            changed_commit = new_branch.commit
        if new_branch.commit != changed_commit:
            changed = True
    for index, instruction in enumerate(new_branch.child_branches):
        child_branch = instruction.recipe_branch
        if_changed_child = None
        if if_changed_from is not None:
            if_changed_child = (
                if_changed_from.child_branches[index].recipe_branch)
        if child_branch is not None:
            child_changed = _resolve_revisions_recurse(
                child_branch, substitute_branch_vars,
                if_changed_from=if_changed_child)
            if child_changed:
                changed = child_changed
    return changed


def resolve_revisions(base_branch, if_changed_from=None,
                      substitute_branch_vars=None):
    """Resolve all the unknowns in `base_branch`.

    This walks the `RecipeBranch` and calls `substitute_branch_vars` for
    each child branch.

    If `if_changed_from` is not None then it should be a second
    `RecipeBranch` to compare `base_branch` against.  If the shape or the
    commit IDs differ then the function will return True.

    :param base_branch: The `RecipeBranch` we plan to build.
    :param if_changed_from: The `RecipeBranch` that we want to compare
        against.
    :param substitute_branch_vars: Callable called for each `RecipeBranch`.
    :return: False if `if_changed_from` is not None, and the shape and
        commits of the two branches are the same.  True otherwise.
    """
    changed = False
    if if_changed_from is not None:
        changed = base_branch.different_shape_to(if_changed_from)
    if_changed_from_revisions = None if changed else if_changed_from

    if substitute_branch_vars is not None:
        real_substitute_branch_vars = (
            lambda child_branch: substitute_branch_vars(
                base_branch, child_branch))
    else:
        real_substitute_branch_vars = None
    changed_revisions = _resolve_revisions_recurse(
        base_branch, real_substitute_branch_vars,
        if_changed_from=if_changed_from_revisions)
    if not changed:
        changed = changed_revisions
    if if_changed_from is not None and not changed:
        return False
    return True


def build_tree(base_branch, target_path):
    logging.info("Building tree.")
    _build_inner_tree(base_branch, target_path)


class ChildBranch:
    """A child branch in a recipe.

    If the nest path is not None it is the path relative to the recipe
    branch where the child branch should be placed.  If it is None then the
    child branch should be merged instead of nested.
    """

    can_have_children = False

    def __init__(self, recipe_branch, nest_path=None):
        self.recipe_branch = recipe_branch
        self.nest_path = nest_path

    def apply(self, target_path):
        raise NotImplementedError(self.apply)

    def as_tuple(self):
        return (self.recipe_branch, self.nest_path)

    def _get_commit_part(self):
        if self.recipe_branch.commit is not None:
            return " git-commit:%s" % self.recipe_branch.commit
        elif self.recipe_branch.revspec is not None:
            return " %s" % self.recipe_branch.revspec
        else:
            return ""

    def __repr__(self):
        return "<%s %r>" % (self.__class__.__name__, self.nest_path)


class CommandInstruction(ChildBranch):

    def apply(self, target_path):
        # it's a command
        logging.info("Running '%s' in '%s'." % (self.nest_path, target_path))
        subprocess.check_call(
            self.nest_path, cwd=target_path, shell=True, stdin=subprocess.PIPE)

    def as_text(self):
        return "%s %s" % (RUN_INSTRUCTION, self.nest_path)


class MergeInstruction(ChildBranch):

    def apply(self, target_path):
        revision_of = ""
        if self.recipe_branch.revspec is not None:
            revision_of = "revision '%s' of " % self.recipe_branch.revspec
        logging.info(
            "Merging %s'%s' in to '%s'." %
            (revision_of, self.recipe_branch.url, target_path))
        merge_branch(self.recipe_branch, target_path)

    def as_text(self):
        commit_part = self._get_commit_part()
        return "%s %s %s%s" % (
            MERGE_INSTRUCTION, self.recipe_branch.name,
            self.recipe_branch.url, commit_part)

    def __repr__(self):
        return "<%s %r>" % (self.__class__.__name__, self.recipe_branch.name)


class NestPartInstruction(ChildBranch):

    def __init__(self, recipe_branch, subpath, target_subdir):
        ChildBranch.__init__(self, recipe_branch)
        self.subpath = subpath
        self.target_subdir = target_subdir

    def apply(self, target_path):
        nest_part_branch(
            self.recipe_branch, target_path, self.subpath,
            target_subdir=self.target_subdir)

    def as_text(self):
        commit_part = self._get_commit_part()
        if commit_part:
            target_subdir = self.target_subdir
            if target_subdir is None:
                target_subdir = self.subpath
            target_commit_part = " %s%s" % (target_subdir, commit_part)
        elif self.target_subdir is not None:
            target_commit_part = " %s" % self.target_subdir
        else:
            target_commit_part = ""
        return "%s %s %s %s%s" % (
            NEST_PART_INSTRUCTION, self.recipe_branch.name,
            self.recipe_branch.url, self.subpath, target_commit_part)


class NestInstruction(ChildBranch):

    can_have_children = True

    def apply(self, target_path):
        _build_inner_tree(
            self.recipe_branch,
            target_path=os.path.join(target_path, self.nest_path))

    def as_text(self):
        commit_part = self._get_commit_part()
        return "%s %s %s %s%s" % (
            NEST_INSTRUCTION, self.recipe_branch.name,
            self.recipe_branch.url, self.nest_path, commit_part)

    def __repr__(self):
        return "<%s %r>" % (self.__class__.__name__,
            self.recipe_branch.name)


class RecipeBranch:
    """A nested structure that represents a Recipe.

    A RecipeBranch has a name and a url (the name can be None for the root
    branch), and optionally child branches that are either merged or nested.

    The `child_branches` attribute is a list of tuples of `ChildBranch`
    objects.  The commit attribute records the commit that the `url` and
    `revspec` resolved to when the `RecipeBranch` was built, or None if it
    has not been built.

    :ivar commit: After this recipe branch has been built, this is set to
        the commit object name that was merged/nested from the branch at
        `self.url`.
    """

    def __init__(self, name, url, revspec=None):
        """Create a `RecipeBranch`.

        :param name: The name for the branch, or None if it is the root.
        :param url: The URL from which to retrieve the branch.
        :param revspec: A revision (as in `git rev-parse`) for the commit to
            use, or None (the default) to use `HEAD`.
        """
        self.name = name
        self.url = url
        self.revspec = revspec
        self.child_branches = []
        self.commit = None
        self.path = None

    @property
    def remote_name(self):
        if self.name is None:
            return 'source'
        return 'source-%s' % self.name

    def _get_git_path(self):
        if self.path is not None:
            return self.path
        elif not urlparse(self.url).scheme:
            return self.url
        else:
            # XXX more specific exception type
            raise Exception(
                "Repository at %s has not been cloned yet" % self.url)

    def git_call(self, *args, **kwargs):
        cmd = ["git", "-C", self._get_git_path()] + list(args)
        silent = kwargs.pop("silent", False)
        # This allows any raised CalledProcessError to contain the output.
        # Writing to stdout/stderr depending on return code is an
        # approximation, but preserves interleaving and should do the job
        # well enough.
        try:
            output = subprocess.check_output(
                cmd, stderr=subprocess.STDOUT, universal_newlines=True,
                **kwargs)
            if silent:
                sys.stdout.write(output)
        except subprocess.CalledProcessError as e:
            if silent:
                sys.stderr.write(e.output)
            raise

    def git_output(self, *args, **kwargs):
        return subprocess.check_output(
            ["git", "-C", self._get_git_path()] + list(args),
            universal_newlines=True, **kwargs)

    def get_revspec(self):
        return self.revspec if self.revspec is not None else "HEAD"

    def resolve_commit(self):
        """Resolve the commit for this branch."""
        # Capturing stderr is a bit dodgy, but it's the most convenient way
        # to capture it for any exceptions.  We know that git rev-parse does
        # not write to stderr on success.
        try:
            self.commit = self.git_output(
                "rev-parse", "%s/%s" % (self.remote_name, self.get_revspec()),
                stderr=subprocess.STDOUT).rstrip("\n")
            return
        except subprocess.CalledProcessError:
            pass
        # Not a remote-prefixed ref. Try a global search.
        # XXX: This allows cross-branch pollution, but we don't have
        # much choice without reimplementing rev-parse ourselves.
        self.commit = self.git_output(
            "rev-parse", self.get_revspec(),
            stderr=subprocess.STDOUT).rstrip("\n")

    def merge_branch(self, branch):
        """Merge a child branch into this one.

        :param branch: The `RecipeBranch` to merge.
        """
        self.child_branches.append(MergeInstruction(branch))

    def nest_part_branch(self, branch, subpath=None, target_subdir=None):
        """Merge subdir of a child branch into this one.

        :param branch: The `RecipeBranch` to merge.
        :param subpath: Only merge files from `branch` that are from this
            path.  For example, subpath="/debian" will only merge changes
            from that directory.
        :param target_subdir: Optional directory in the target tree to merge
            that subpath into.  Defaults to the basename of `subpath`.
        """
        self.child_branches.append(
            NestPartInstruction(branch, subpath, target_subdir))

    def nest_branch(self, location, branch):
        """Nest a child branch into this one.

        :param location: The relative path at which this branch should be
            nested.
        :param branch: The `RecipeBranch` to nest.
        """
        if any(location == b.nest_path for b in self.child_branches):
            raise AssertionError(
                "%s already has a branch nested there" % location)
        self.child_branches.append(NestInstruction(branch, location))

    def run_command(self, command):
        """Set a command to be run.

        :param command: The command to be run.
        """
        self.child_branches.append(CommandInstruction(None, command))

    def different_shape_to(self, other_branch):
        """Test whether the name, url, and child_branches are the same."""
        if self.name != other_branch.name:
            return True
        if self.url != other_branch.url:
            return True
        if len(self.child_branches) != len(other_branch.child_branches):
            return True
        for index, instruction in enumerate(self.child_branches):
            child_branch = instruction.recipe_branch
            nest_location = instruction.nest_path
            other_instruction = other_branch.child_branches[index]
            other_child_branch = other_instruction.recipe_branch
            other_nest_location = other_instruction.nest_path
            if nest_location != other_nest_location:
                return True
            if ((child_branch is None and other_child_branch is not None) or
                    (child_branch is not None and other_child_branch is None)):
                return True
            # If child_branch is None then other_child_branch must be None
            # too, meaning that they are both run instructions.  We would
            # compare their nest locations (commands), but that has already
            # been done, so just guard.
            if (child_branch is not None and
                    child_branch.different_shape_to(other_child_branch)):
                return True
        return False

    def iter_all_instructions(self):
        """Iter over all instructions under this branch."""
        for instruction in self.child_branches:
            yield instruction
            child_branch = instruction.recipe_branch
            if child_branch is None:
                continue
            for instruction in child_branch.iter_all_instructions():
                yield instruction

    def iter_all_branches(self):
        """Iterate over all branches."""
        yield self
        for instruction in self.child_branches:
            child_branch = instruction.recipe_branch
            if child_branch is None:
                continue
            for subbranch in child_branch.iter_all_branches():
                yield subbranch

    def lookup_branch(self, name):
        """Look up a branch by its name."""
        for branch in self.iter_all_branches():
            if branch.name == name:
                return branch
        else:
            raise KeyError(name)

    def list_branch_names(self):
        """List all of the branch names under this one.

        :return: A list of the branch names.
        :rtype: list(str)
        """
        return [
            branch.name for branch in self.iter_all_branches()
            if branch.name is not None]

    def __repr__(self):
        return "<%s %r>" % (self.__class__.__name__, self.name)


class BaseRecipeBranch(RecipeBranch):
    """The `RecipeBranch` that is at the root of a recipe."""

    def __init__(self, url, deb_version, format_version, revspec=None):
        """Create a `BaseRecipeBranch`.

        :param deb_version: The template to use for the version number.
            Should be None for anything except the root branch.
        """
        super().__init__(None, url, revspec=revspec)
        self.deb_version = deb_version
        self.format_version = format_version
        if not urlparse(url).scheme:
            self.path = url

    def _add_child_branches_to_manifest(self, child_branches, indent_level):
        manifest = ""
        for instruction in child_branches:
            manifest += "%s%s\n" % ("  " * indent_level, instruction.as_text())
            if instruction.can_have_children:
                manifest += self._add_child_branches_to_manifest(
                    instruction.recipe_branch.child_branches, indent_level + 1)
        return manifest

    def __str__(self):
        return self.get_recipe_text(validate=True)

    def get_recipe_text(self, validate=False):
        manifest = "# git-build-recipe format %s" % str(self.format_version)
        if self.deb_version is not None:
            # TODO: should we store the expanded version that was used?
            manifest += " deb-version %s" % self.deb_version
        manifest += "\n"
        if self.commit is not None:
            manifest += "%s git-commit:%s\n" % (self.url, self.commit)
        elif self.revspec is not None:
            manifest += "%s %s\n" % (self.url, self.revspec)
        else:
            manifest += "%s\n" % self.url
        manifest += self._add_child_branches_to_manifest(
            self.child_branches, 0)
        if validate:
            # Sanity check.
            # TODO: write a function that compares the result of this parse
            # with the branch that we built it from.
            RecipeParser(manifest).parse()
        return manifest


class RecipeParseError(FormattedError):

    _fmt = "Error parsing %(filename)s:%(line)s:%(char)s: %(problem)s."

    def __init__(self, filename, line, char, problem):
        super().__init__(
            filename=filename, line=line, char=char, problem=problem)


class InstructionParseError(RecipeParseError):

    _fmt = RecipeParseError._fmt + "\nUsage: %(usage)s"

    def __init__(self, filename, line, char, problem, instruction):
        super().__init__(filename, line, char, problem)
        self.usage = USAGE[instruction]


class ForbiddenInstructionError(RecipeParseError):

    def __init__(self, filename, line, char, problem, instruction_name=None):
        super().__init__(filename, line, char, problem)
        self.instruction_name = instruction_name


class RecipeParser:
    """Parse a recipe.

    The parse() method is probably the only one that interests you.
    """

    whitespace_chars = " \t"
    eol_char = "\n"
    digit_chars = ("0", "1", "2", "3", "4", "5", "6", "7", "8", "9")

    NEWEST_VERSION = 0.4

    def __init__(self, f):
        """Create a `RecipeParser`.

        :param f: Either the recipe as a string, or a file-like object to
            take it from.
        """
        if getattr(f, "read", None) is not None:
            self.text = f.read()
        else:
            self.text = f
        self.filename = None
        if hasattr(f, "name"):
            self.filename = f.name
        if self.filename is None:
            self.filename = "recipe"

    def parse(self, permitted_instructions=None):
        """Parse the recipe.

        :param permitted_instructions: A list of instructions that you want
            to allow.  Defaults to None allowing them all.
        :type permitted_instructions: list(str) or None
        :return: A `RecipeBranch` representing the recipe.
        """
        self.lines = self.text.split("\n")
        self.index = 0
        self.line_index = 0
        self.current_line = self.lines[self.line_index]
        self.current_indent_level = 0
        self.seen_nicks = set()
        self.seen_paths = {".": 1}
        version, deb_version = self.parse_header()
        self.version = version
        last_instruction = None
        active_branches = []
        last_branch = None
        while self.line_index < len(self.lines):
            if self.is_blankline():
                self.new_line()
                continue
            comment = self.parse_comment_line()
            if comment is not None:
                self.new_line()
                continue
            old_indent_level = self.parse_indent()
            if old_indent_level is not None:
                if (old_indent_level < self.current_indent_level
                    and last_instruction != NEST_INSTRUCTION):
                    self.throw_parse_error(
                        "Not allowed to indent unless after a '%s' line" %
                        NEST_INSTRUCTION)
                if old_indent_level < self.current_indent_level:
                    active_branches.append(last_branch)
                else:
                    unindent = self.current_indent_level - old_indent_level
                    active_branches = active_branches[:unindent]
            if last_instruction is None:
                url = self.take_to_whitespace("branch to start from")
                revspec = self.parse_optional_revspec()
                self.new_line()
                last_branch = BaseRecipeBranch(
                    url, deb_version, self.version, revspec=revspec)
                active_branches = [last_branch]
                last_instruction = ""
            else:
                instruction = self.parse_instruction(
                    permitted_instructions=permitted_instructions)
                if instruction == RUN_INSTRUCTION:
                    self.parse_whitespace(
                        "the command", instruction=instruction)
                    command = self.take_to_newline().strip()
                    self.new_line()
                    active_branches[-1].run_command(command)
                else:
                    branch_id = self.parse_branch_id(instruction)
                    url = self.parse_branch_url(instruction)
                    if instruction == NEST_INSTRUCTION:
                        location = self.parse_branch_location(instruction)
                    if instruction == NEST_PART_INSTRUCTION:
                        path = self.parse_subpath(instruction)
                        target_subdir = self.parse_optional_path()
                        if target_subdir == "":
                            target_subdir = None
                            revspec = None
                        else:
                            revspec = self.parse_optional_revspec()
                    else:
                        revspec = self.parse_optional_revspec()
                    self.new_line(instruction)
                    last_branch = RecipeBranch(branch_id, url, revspec=revspec)
                    if instruction == NEST_INSTRUCTION:
                        active_branches[-1].nest_branch(location, last_branch)
                    elif instruction == MERGE_INSTRUCTION:
                        active_branches[-1].merge_branch(last_branch)
                    elif instruction == NEST_PART_INSTRUCTION:
                        active_branches[-1].nest_part_branch(
                            last_branch, path, target_subdir)
                last_instruction = instruction
        if len(active_branches) == 0:
            self.throw_parse_error("Empty recipe")
        return active_branches[0]

    def parse_header(self):
        self.parse_char("#")
        # Try git-build-recipe last even though it's now the preferred form,
        # in order that user-visible error messages look sensible.
        try:
            self.parse_word("bzr-builder", require_whitespace=False)
        except RecipeParseError:
            self.parse_word("git-build-recipe", require_whitespace=False)
        self.parse_word("format")
        version, version_str = self.peek_float("format version")
        if version > self.NEWEST_VERSION:
            self.throw_parse_error("Unknown format: '%s'" % str(version))
        self.take_chars(len(version_str))
        deb_version = self.parse_optional_deb_version()
        self.new_line()
        return version, deb_version

    def parse_instruction(self, permitted_instructions=None):
        if self.version < 0.2:
            options = (MERGE_INSTRUCTION, NEST_INSTRUCTION)
            options_str = "'%s' or '%s'" % options
        elif self.version < 0.3:
            options = (MERGE_INSTRUCTION, NEST_INSTRUCTION, RUN_INSTRUCTION)
            options_str = "'%s', '%s', or '%s'" % options
        else:
            options = (
                MERGE_INSTRUCTION, NEST_INSTRUCTION, NEST_PART_INSTRUCTION,
                RUN_INSTRUCTION)
            options_str = "'%s', '%s', '%s', or '%s'" % options
        instruction = self.peek_to_whitespace()
        if instruction is None:
            self.throw_parse_error(
                "End of line while looking for %s" % options_str)
        if instruction in options:
            if permitted_instructions is not None:
                if instruction not in permitted_instructions:
                    self.throw_parse_error(
                        "The '%s' instruction is forbidden" % instruction,
                        cls=ForbiddenInstructionError,
                        instruction_name=instruction)
            self.take_chars(len(instruction))
            return instruction
        self.throw_parse_error(
            "Expecting %s, got '%s'" % (options_str, instruction))

    def parse_branch_id(self, instruction):
        self.parse_whitespace("the branch id", instruction=instruction)
        branch_id = self.peek_to_whitespace()
        if branch_id is None:
            self.throw_parse_error(
                "End of line while looking for the branch id",
                cls=InstructionParseError, instruction=instruction)
        if branch_id in self.seen_nicks:
            self.throw_parse_error(
                "'%s' was already used to identify a branch." % branch_id)
        self.take_chars(len(branch_id))
        self.seen_nicks.add(branch_id)
        return branch_id

    def parse_branch_url(self, instruction):
        self.parse_whitespace("the branch url", instruction=instruction)
        return self.take_to_whitespace("the branch url", instruction)

    def parse_branch_location(self, instruction):
        # FIXME: Needs a better term
        self.parse_whitespace("the location to nest")
        location = self.peek_to_whitespace()
        if location is None:
            self.throw_parse_error(
                "End of line while looking for the location to nest",
                cls=InstructionParseError, instruction=instruction)
        norm_location = os.path.normpath(location)
        if norm_location in self.seen_paths:
            self.throw_parse_error(
                "The path '%s' is a duplicate of the one used on line %d." % (
                    location, self.seen_paths[norm_location]),
                cls=InstructionParseError, instruction=instruction)
        if os.path.isabs(norm_location):
            self.throw_parse_error(
                "Absolute paths are not allowed: %s" % location,
                cls=InstructionParseError, instruction=instruction)
        if norm_location.startswith(".."):
            self.throw_parse_error(
                "Paths outside the current directory are not allowed: %s" % (
                    location),
                cls=InstructionParseError, instruction=instruction)
        self.take_chars(len(location))
        self.seen_paths[norm_location] = self.line_index + 1
        return location

    def parse_subpath(self, instruction):
        self.parse_whitespace("the subpath to merge", instruction=instruction)
        return self.take_to_whitespace("the subpath to merge", instruction)

    def parse_revspec(self):
        self.parse_whitespace("the revspec")
        return self.take_to_whitespace("the revspec")

    def parse_optional_deb_version(self):
        self.parse_whitespace("'deb-version'", require=False)
        actual = self.peek_to_whitespace()
        if actual is None:
            # End of line, no version
            return None
        if actual != "deb-version":
            self.throw_expecting_error("deb-version", actual)
        self.take_chars(len("deb-version"))
        self.parse_whitespace("a value for 'deb-version'")
        return self.take_to_whitespace("a value for 'deb-version'")

    def parse_optional_revspec(self):
        self.parse_whitespace(None, require=False)
        revspec = self.peek_to_whitespace()
        if revspec is not None:
            self.take_chars(len(revspec))
        return revspec

    def parse_optional_path(self):
        self.parse_whitespace(None, require=False)
        path = self.peek_to_whitespace()
        if path is not None:
            self.take_chars(len(path))
        return path

    def throw_parse_error(self, problem, cls=None, **kwargs):
        if cls is None:
            cls = RecipeParseError
        raise cls(
            self.filename, self.line_index + 1, self.index + 1, problem,
            **kwargs)

    def throw_expecting_error(self, expected, actual):
        self.throw_parse_error("Expecting '%s', got '%s'" % (expected, actual))

    def throw_eol(self, expected):
        self.throw_parse_error("End of line while looking for '%s'" % expected)

    def new_line(self, instruction=None):
        # Jump over any whitespace
        self.parse_whitespace(None, require=False)
        remaining = self.peek_to_whitespace()
        if remaining is not None:
            kwargs = {}
            if instruction is not None:
                kwargs = {
                    "cls": InstructionParseError,
                    "instruction": instruction,
                    }
            self.throw_parse_error(
                "Expecting the end of the line, got '%s'" % remaining,
                **kwargs)
        self.index = 0
        self.line_index += 1
        if self.line_index >= len(self.lines):
            self.current_line = None
        else:
            self.current_line = self.lines[self.line_index]

    def is_blankline(self):
        whitespace = self.peek_whitespace()
        if whitespace is None:
            return True
        return self.peek_char(skip=len(whitespace)) is None

    def take_char(self):
        if self.index >= len(self.current_line):
            return None
        self.index += 1
        return self.current_line[self.index-1]

    def take_chars(self, num):
        ret = ""
        for i in range(num):
            char = self.take_char()
            if char is None:
                return None
            ret += char
        return ret

    def peek_char(self, skip=0):
        if self.index + skip >= len(self.current_line):
            return None
        return self.current_line[self.index + skip]

    def parse_char(self, char):
        actual = self.peek_char()
        if actual is None:
            self.throw_eol(char)
        if actual == char:
            self.take_char()
            return char
        self.throw_expecting_error(char, actual)

    def parse_indent(self):
        """Parse the indent from the start of the line."""
        # FIXME: should just peek the whitespace
        new_indent = self.parse_whitespace(None, require=False)
        # FIXME: These checks should probably come after we check whether
        # any change in indent is legal at this point:
        # "Indents of 3 spaces aren't allowed" -> make it 2 spaces
        # -> "oh, you aren't allowed to indent at that point anyway"
        if "\t" in new_indent:
            self.throw_parse_error("Indents may not be done by tabs")
        if (len(new_indent) % 2 != 0):
            self.throw_parse_error("Indent not a multiple of two spaces")
        new_indent_level = len(new_indent) // 2
        if new_indent_level != self.current_indent_level:
            old_indent_level = self.current_indent_level
            self.current_indent_level = new_indent_level
            if (new_indent_level > old_indent_level and
                    new_indent_level - old_indent_level != 1):
               self.throw_parse_error(
                    "Indented by more than two spaces at once")
            return old_indent_level
        return None

    def parse_whitespace(self, looking_for, require=True, instruction=None):
        if require:
            actual = self.peek_char()
            if actual is None:
                kwargs = {}
                if instruction is not None:
                    kwargs = {
                        "cls": InstructionParseError,
                        "instruction": instruction,
                        }
                self.throw_parse_error(
                    "End of line while looking for %s" % looking_for, **kwargs)
            if actual not in self.whitespace_chars:
                self.throw_parse_error(
                    "Expecting whitespace before %s, got '%s'." %
                    (looking_for, actual))
        ret = ""
        actual = self.peek_char()
        while actual is not None and actual in self.whitespace_chars:
            self.take_char()
            ret += actual
            actual = self.peek_char()
        return ret

    def peek_whitespace(self):
        ret = ""
        char = self.peek_char()
        if char is None:
            return char
        count = 0
        while char is not None and char in self.whitespace_chars:
            ret += char
            count += 1
            char = self.peek_char(skip=count)
        return ret

    def parse_word(self, expected, require_whitespace=True):
        self.parse_whitespace("'%s'" % expected, require=require_whitespace)
        length = len(expected)
        actual = self.peek_to_whitespace()
        if actual == expected:
            self.take_chars(length)
            return expected
        if actual is None:
            self.throw_eol(expected)
        self.throw_expecting_error(expected, actual)

    def peek_to_whitespace(self):
        ret = ""
        char = self.peek_char()
        if char is None:
            return char
        count = 0
        while char is not None and char not in self.whitespace_chars:
            ret += char
            count += 1
            char = self.peek_char(skip=count)
        return ret

    def take_to_whitespace(self, looking_for, instruction=None):
        text = self.peek_to_whitespace()
        if text is None:
            kwargs = {}
            if instruction is not None:
                kwargs = {
                    "cls": InstructionParseError,
                    "instruction": instruction,
                    }
            self.throw_parse_error(
                "End of line while looking for %s" % looking_for, **kwargs)
        self.take_chars(len(text))
        return text

    def peek_float(self, looking_for):
        self.parse_whitespace(looking_for)
        ret = self._parse_integer()
        conv_fn = int
        if ret == "":
            self.throw_parse_error(
                "Expecting a float, got '%s'" % self.peek_to_whitespace())
        if self.peek_char(skip=len(ret)) == ".":
            conv_fn = float
            ret2 = self._parse_integer(skip=(len(ret) + 1))
            if ret2 == "":
                self.throw_parse_error(
                    "Expecting a float, got '%s'" % self.peek_to_whitespace())
            ret += "." + ret2
        try:
            fl = conv_fn(ret)
        except ValueError:
            self.throw_parse_error("Expecting a float, got '%s'" % ret)
        return (fl, ret)

    def _parse_integer(self, skip=0):
        i = skip
        ret = ""
        while True:
            char = self.peek_char(skip=i)
            if char not in self.digit_chars:
                break
            ret += char
            i += 1
        return ret

    def parse_integer(self):
        ret = self._parse_integer()
        if ret == "":
            self.throw_parse_error(
                "Expected an integer, found %s" % self.peek_to_whitespace())
        self.take_chars(len(ret))
        return ret

    def take_to_newline(self):
        text = self.current_line[self.index:]
        self.index += len(text)
        return text

    def parse_comment_line(self):
        whitespace = self.peek_whitespace()
        if whitespace is None:
            return ""
        if self.peek_char(skip=len(whitespace)) is None:
            return ""
        if self.peek_char(skip=len(whitespace)) != "#":
            return None
        self.parse_whitespace(None, require=False)
        comment = self.current_line[self.index:]
        self.index += len(comment)
        return comment


def parse_recipe(recipe_file, safe=False):
    parser = RecipeParser(recipe_file)
    if safe:
        permitted_instructions = SAFE_INSTRUCTIONS
    else:
        permitted_instructions = None
    return parser.parse(permitted_instructions=permitted_instructions)
